#! /usr/bin/env python
# -*- coding: utf-8 -*-

# part of module
from .collocates import add_ams
# requirements
from pandas import DataFrame, MultiIndex
# logging
import logging
logger = logging.getLogger(__name__)


class Keywords:
    """ keyword analysis """

    def __init__(self, corpus, df_dump, p_query=['lemma']):

        self.corpus = corpus

        # activate dump
        self.size = len(df_dump)

        # consistency check
        if self.size == 0:
            logger.warning('cannot calculate keywords on 0 regions')
            return

        # determine layer to work on
        self.p_query = [p_query] if isinstance(p_query, str) else p_query

        # TODO: also comprises s-att -- implement convenient retrieval
        p_available = set(self.corpus.attributes_available['attribute'].values)
        if not set(self.p_query).issubset(p_available):
            logger.warning(
                'specfied p-attribute(s) (%s) not available' % " ".join(self.p_query)
            )
            logger.warning('falling back to primary layer')
            self.p_query = ['word']

        # collect context and save result
        logger.info('collecting token counts of subcorpus')
        counts = corpus.counts.dump(
            df_dump, start='match', end='matchend', p_atts=self.p_query, split=True
        )
        counts.columns = ['f']

        self.counts = counts

    def show(self, order='f', cut_off=100, ams=None,
             min_freq=2, frequencies=True, flags=None,
             marginals='corpus'):

        # consistency check
        if self.counts.empty:
            logger.warning("nothing to show")
            return DataFrame()

        # get frequencies
        f = self.counts.loc[self.counts['f'] >= min_freq]

        # determine reference frequency
        if isinstance(marginals, str):
            N = self.corpus.corpus_size
            # get marginals
            if marginals == 'corpus':
                if len(self.p_query) == 1:
                    # coerce to multiindex (what was I thinking?)
                    f2 = self.corpus.marginals(
                        f.index.get_level_values(self.p_query[0]), self.p_query[0]
                    )
                    f2.index = MultiIndex.from_tuples(
                        f2.index.map(lambda x: (x, )), names=self.p_query
                    )
                else:
                    f2 = self.corpus.marginals_complex(f.index, self.p_query)
                f2.columns = ['f2']
            else:
                raise NotImplementedError

        elif isinstance(marginals, DataFrame):
            f2 = marginals
            f2.columns = ['f2']
            N = f2['f2'].sum()

        else:
            raise NotImplementedError

        # get sub-corpus size
        f1 = self.counts['f'].sum()

        keywords = add_ams(
            f, f1, f2, N,
            min_freq, order, cut_off, flags, ams, frequencies
        )

        # deal with index
        if len(self.p_query) == 1:
            keywords.index = keywords.index.map(lambda x: x[0])
            keywords.index.name = self.p_query[0]

        return keywords



from ccc import Corpus
from ccc.discoursemes import Discourseme, DiscoursemeConstellation
from collections import defaultdict
import pytest


from .conftest import DATA_PATH


# backend.analysis.ccc methods ####################################
def get_concordance(corpus_name, topic_items, topic_name, s_context,
                    window_size, context=20,
                    additional_discoursemes={}, p_query='lemma',
                    p_show=['word', 'lemma'], s_show=['text_id'],
                    s_query=None, order='random', cut_off=100,
                    form='dataframe'):

    if s_query is None:
        s_query = s_context

    # init corpus
    corpus = Corpus(corpus_name)

    # init topic discourseme
    topic_disc = Discourseme(
        corpus,
        items=topic_items,
        p_query=p_query,
        s_query=s_query
    )

    # init discourseme constellation
    dp = DiscoursemeConstellation(topic_disc)

    # id2name: mapper from ccc cache names to discourseme names
    id2name = {
        topic_disc.idx: topic_name
    }

    # add discoursemes to discourseme constellation
    for key in additional_discoursemes.keys():
        idx = dp.add_items(additional_discoursemes[key])
        id2name[idx] = key

    # extract concordance
    concordance = dp.concordance(
        window=window_size,
        matches=None,
        p_show=p_show,
        s_show=s_show,
        order=order,
        cut_off=cut_off,
        form=form
    )

    if concordance.empty:
        return None

    # convert each concordance line to dictionary, create roles
    # TODO: implement as form='json' in cwb-ccc
    concordance = concordance.reset_index()
    ret = dict()
    for idx, df in zip(concordance['match'], concordance['dataframe']):

        # rename columns according to given names for discoursemes
        df = df.rename(columns=id2name)

        # get roles as dict
        roles = defaultdict(list)
        for cpos, row in df.iterrows():
            for col_name in id2name.values():
                if row[col_name]:
                    # TODO: propagate proper info about discourseme names
                    if col_name == topic_name:
                        roles[cpos].append('topic')
                    else:
                        roles[cpos].append('collocate')
                else:
                    roles[cpos].append(None)

        ret[idx] = {
            'word': list(df['word']),    # list of words
            p_query: list(df[p_query]),  # secondary p-att
            'role': list(roles.values())  # roles
        }

    return ret


def get_collocates(corpus_name, topic_items, s_context, window_sizes,
                   context=20, additional_discoursemes=[],
                   p_query='lemma', s_query=None, ams=None,
                   cut_off=200, order='log_likelihood'):

    if s_query is None:
        s_query = s_context

    corpus = Corpus(corpus_name)

    topic_disc = Discourseme(
        corpus,
        items=topic_items,
        p_query=p_query,
        s_query=s_query
    )

    # TODO speed up in backend
    collocates = dict()
    for window in window_sizes:

        if not additional_discoursemes:
            # single discourseme
            coll_window = topic_disc.collocates(
                window_sizes=window,
                order=order,
                cut_off=cut_off,
                p_query=p_query,
                ams=ams,
                min_freq=2,
                frequencies=False,
                flags=None
            )
        else:
            # discursive position
            dp = DiscoursemeConstellation(topic_disc)
            for key in additional_discoursemes.keys():
                dp.add_items(additional_discoursemes[key])

            coll_window = dp.collocates(
                window=window,
                order=order,
                cut_off=cut_off,
                p_query=p_query,
                ams=ams,
                min_freq=2,
                frequencies=False,
                flags=None
            )

        # drop superfluous columns and sort
        coll_window = coll_window[[
            'log_likelihood',
            'log_ratio',
            'f',
            'f2',
            'mutual_information',
            'z_score',
            't_score'
        ]]

        # rename AMs
        am_dict = {
            'log_likelihood': 'log-likelihood',
            'f': 'co-oc. freq.',
            'mutual_information': 'mutual information',
            'log_ratio': 'log-ratio',
            'f2': 'marginal freq.',
            't_score': 't-score',
            'z_score': 'z-score'
        }
        collocates[window] = coll_window.rename(am_dict, axis=1)

    return collocates


def get_corpus(corpus_settings, data_path=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_path=corpus_settings['registry_path'],
        lib_path=corpus_settings.get('lib_path', None),
        data_path=data_path
    )


@pytest.mark.discourseme
def test_init(germaparl):
    corpus = get_corpus(germaparl)
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma', 's',
    )
    assert(len(topic.dump.df) == 1640)


@pytest.mark.discourseme
def test_concordance(germaparl):
    corpus = get_corpus(germaparl)
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma', 's'
    )
    print(topic.concordance())
    print(topic.concordance(10))
    print(topic.concordance(30))


@pytest.mark.discourseme
def test_disc_concordance_form(germaparl):

    corpus = get_corpus(germaparl)
    # init topic disc
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's'
    )
    print(topic.concordance(cut_off=None, form='kwic'))
    print(topic.concordance(matches=[148430], cut_off=None, form='slots'))


@pytest.mark.discourseme
def test_disc_collocates(germaparl):

    corpus = get_corpus(germaparl)

    # init topic disc
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's'
    )
    print(topic.collocates())


# constellations
@pytest.mark.disccon
def test_disccon(germaparl):

    corpus = get_corpus(germaparl)

    # init topic disc
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's'
    )

    # two floating discoursemes
    disc1 = Discourseme(
        corpus,
        ["sollen", "müssen"],
        'lemma',
        's'
    )
    disc2 = Discourseme(
        corpus,
        ["und"],
        'lemma',
        's'
    )
    discon = DiscoursemeConstellation(topic, [disc1, disc2])
    discon.slice_discs(10)
    print(discon.df_nodes[10])
    print(discon.df_nodes[10].columns)
    print(discon.discoursemes)


@pytest.mark.disccon
def test_disccon_2(germaparl):

    corpus = get_corpus(germaparl)

    # init topic disc
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's',
    )
    discon = DiscoursemeConstellation(topic)
    # two floating discoursemes
    discon.add_items(["sollen", "müssen"])
    discon.add_items(["und"])
    discon.slice_discs(10)
    print(discon.df_nodes[10].keys())
    print(discon.df_nodes[10])


@pytest.mark.disccon
def test_disccon_concordance(germaparl):

    corpus = get_corpus(germaparl)

    # three discoursemes
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's'
    )
    disc1 = Discourseme(
        corpus,
        ["sollen", "müssen"],
        'lemma',
        's'
    )
    disc2 = Discourseme(
        corpus,
        ["und"],
        'lemma',
        's'
    )

    # init discursive position
    disccon = DiscoursemeConstellation(topic, [disc1, disc2])
    # show concordance
    print(disccon.concordance())
    print(disccon.concordance(p_show=['word', 'lemma'])['dataframe'].iloc[0])


@pytest.mark.disccon
def test_disccon_collocates(germaparl):

    corpus = get_corpus(germaparl)

    # three discoursemes
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's',
        's'
    )
    disc1 = Discourseme(
        corpus,
        ["sollen", "müssen", "machen"],
        'lemma',
        's',
        's'
    )
    disc2 = Discourseme(
        corpus,
        ["und"],
        'lemma',
        's',
        's'
    )

    # init discursive position
    discpos = DiscoursemeConstellation(topic, [disc1, disc2])
    # show collocates
    print(discpos.collocates())


@pytest.mark.disccon
def test_disccon_collocates_empty(germaparl):

    corpus = get_corpus(germaparl)
    # three discoursemes
    topic = Discourseme(
        corpus,
        ["SPD", "CSU", "Grünen"],
        'lemma',
        's'
    )
    # init discursive position
    disccon = DiscoursemeConstellation(topic)
    disccon.add_items(["Verhandlung"])
    # show collocates
    assert(disccon.collocates(window=5).empty)


@pytest.mark.disccon
def test_disccon_collocates_nodes(germaparl):
    corpus = get_corpus(germaparl)
    topic = Discourseme(
        corpus,
        [",", ".", ")", "("],
        'lemma',
        's',
        escape=True
    )
    df = topic.collocates(cut_off=None)[5]
    assert("," not in df.index)
    assert("(" not in df.index)

    topic2 = DiscoursemeConstellation(topic)
    df2 = topic2.collocates(cut_off=None)
    assert(df2.equals(df))


@pytest.mark.disccon
def test_disccon_collocates_range(germaparl):

    corpus = get_corpus(germaparl)

    # three discoursemes
    topic = Discourseme(
        corpus,
        [",", ".", ")", "("],
        'lemma',
        's',
        's',
        escape=True
    )
    disc1 = Discourseme(
        corpus,
        ["die", "sie", "und"],
        'lemma',
        's',
        's'
    )
    disc2 = Discourseme(
        corpus,
        ["sein", "in", "eine", "zu", "haben"],
        'lemma',
        's',
        's'
    )
    disccon = DiscoursemeConstellation(topic, [disc1, disc2])
    df = disccon.collocates(cut_off=None)
    print(df)


# @pytest.mark.mmda
def test_get_concordance_simple():

    # init corpus
    corpus_name = "GERMAPARL1318"

    # request parameters
    topic_items = ['Atomkraft', 'Nuklearnergie', 'Atomenergie']
    topic_name = "Atomkraft"
    s_context = "s"
    window_size = 3
    context = 15
    additional_discoursemes = {}
    p_query = 'lemma'
    p_show = ['word', 'lemma']
    s_show = ['s']
    s_query = s_context
    order = 'random'
    cut_off = 100
    form = 'dataframe'

    conc = get_concordance(
        corpus_name,
        topic_items,
        topic_name,
        s_context,
        window_size,
        context,
        additional_discoursemes,
        p_query,
        p_show,
        s_show,
        s_query,
        order,
        cut_off,
        form
    )
    from pprint import pprint
    pprint(conc[list(conc.keys())[0]])


@pytest.mark.mmda
def test_get_collocates_simple():

    # init corpus
    corpus_name = "GERMAPARL1318"

    # request parameters
    topic_items = ['Atomkraft', 'Nuklearnergie', 'Atomenergie']
    s_context = "s"
    window_size = 3
    context = 15
    additional_discoursemes = {}
    p_query = 'lemma'
    s_query = s_context
    order = 'log_likelihood'
    cut_off = 100

    coll = get_collocates(
        corpus_name,
        topic_items,
        s_context,
        [window_size],
        context,
        additional_discoursemes,
        p_query,
        s_query,
        None,
        cut_off,
        order
    )
    from pprint import pprint
    pprint(coll[list(coll.keys())[0]])



from ccc.cwb import Corpus
import pytest

from .conftest import LOCAL, DATA_PATH


def get_corpus(corpus_settings, data_path=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_path=corpus_settings['registry_path'],
        lib_path=corpus_settings.get('lib_path', None),
        data_path=data_path
    )


def test_query2dump(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump)


def test_query2dump_name(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"', name="Test")
    print(dump)


def test_breakdown(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.breakdown())


def test_dump_matches(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.matches())


def test_dump_matches1(germaparl):
    corpus = get_corpus(germaparl)
    dump_base = corpus.query(
        r'[pos="NE"]? [pos="NE"] "\[" ".*" "]")', name="Base"
    )
    tokens_base = len(dump_base.matches())
    corpus.subcorpus = "Base"
    dump_neg = corpus.query('[pos="NE"]')
    tokens_neg = len(dump_neg.matches())
    print(tokens_base - tokens_neg)


def test_dump_context(germaparl):
    corpus = get_corpus(germaparl)

    # init topic disc
    dump = corpus.query('"SPD"')
    print(dump.context())


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
def test_satt2dump(brexit):

    corpus = get_corpus(brexit)
    ids = {
        't740982320711249920',
        't731037753241112576',
        't729363812802039814',
        't733648546881277953',
        't741216447595220992',
        't705780723018539012',
        't745930343627243520',
        't730870826178904065',
        't745691821477605377',
        't730419966818783232',
        't746069538693750784'
    }
    dump = corpus.query_s_att('tweet_id', ids)
    print(dump.concordance())


def test_concordance(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.concordance())


def test_concordance_options(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.concordance(form='raw'))
    print(dump.concordance(form='simple'))
    print(dump.concordance(form='kwic'))
    print(dump.concordance(form='dataframe'))
    print(dump.concordance(form='slots'))


def test_concordance_set_context(germaparl):

    corpus = get_corpus(germaparl)
    dump = corpus.query('"CSU"', context_break='text')
    print(dump)
    dump.set_context(10, context_break='text', context_right=10)
    print(dump)


def test_collocates(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.collocates())


def test_collocates_options(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.collocates(order='log_likelihood', cut_off=200))


def test_keywords(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD" expand to s')
    print(dump.keywords())


def test_keywords_options(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD" expand to s')
    print(dump.keywords(order='log_ratio', cut_off=200))


def test_context_matches(germaparl):
    corpus = get_corpus(germaparl)
    dump = corpus.query('"SPD"')
    print(dump.matches())
    print(dump.context())


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
def test_argmin_query(brexit):
    corpus = get_corpus(brexit)

    query = brexit['query_argmin']

    dump = corpus.query(
        cqp_query=query['cqp'],
        context=query.get('context', None),
        context_break=query.get('s_context', None),
        corrections=query['corrections'],
        match_strategy=query['match_strategy']
    )

    conc = dump.concordance(
        p_show=query['p_show'],
        s_show=query['s_show'],
        slots=query['slots'],
        order='first',
        cut_off=None,
        form='slots'
    )

    print(conc)



from .collocates import Collocates
# logging
import logging
logger = logging.getLogger(__name__)


class UFA:

    def __init__(self, corpus, s_dict, s_att='text_id'):
        """
        :param Corpus corpus: corpus to work on
        :param dict ids: dicitonary of {subcorpus_name: set of values for s_att}
        :param str context_break: s-attribute that is used for definition of regions
        """

        self.corpus = corpus.copy()
        self.s_dict = s_dict
        self.s_att = s_att

        logger.info("creating subcorpora ...")
        dumps = dict()
        i = 0
        for s in self.s_dict.keys():
            i += 1
            logger.info("... subcorpus %d of %d" % (i, len(self.s_dict)))
            dumps[s] = corpus.query_s_att(
                self.s_att, self.s_dict[s]
            )

        self.dumps = dumps

    def keywords(self, p_query='lemma', order='log_likelihood', cut_off=100,
                 ams=None, min_freq=2, frequencies=True, flags=None,
                 subset=None):
        """
        :return: dictionary of {subcorpus_name: table}
        :rtype: dict
        """
        if subset is None:
            subset = list(self.s_dict.keys())

        logger.info("computing keyword tables ...")
        tables = dict()
        i = 0
        for s in subset:
            i += 1
            logger.info("... table %d of %d" % (i, len(subset)))
            dump = self.dumps[s]
            tables[s] = dump.keywords(
                p_query=p_query, order=order, cut_off=cut_off,
                ams=ams, min_freq=min_freq,
                frequencies=frequencies, flags=flags
            )

        return tables

    def collocates(self, cqp_query, window=5, p_query='lemma',
                   order='log_likelihood', cut_off=100, ams=None, min_freq=2,
                   frequencies=True, flags=None, subset=None, context_break=None):
        """
        :return: dictionary of {subcorpus_name: table}
        :rtype: dict
        """
        if subset is None:
            subset = list(self.s_dict.keys())

        tables = dict()

        if context_break is None:
            context_break = self.s_att

        # run query and de-construct dump
        dump_glob = self.corpus.query(
            cqp_query,
            context=window,
            context_break=context_break
        ).df
        df_glob = self.corpus.dump2satt(dump_glob, context_break)

        logger.info("computing collocate tables ...")
        tables = dict()
        i = 0
        for s in subset:
            i += 1
            logger.info("... table %d of %d" % (i, len(subset)))
            dump = self.dumps[s]
            # get local marginals (f2)
            marginals_loc = self.corpus.counts.dump(
                dump.df, split=True, p_atts=[p_query]
            )
            # create collocates table
            df_loc = df_glob.loc[
                df_glob[self.s_att].isin(self.s_dict[s])
            ]
            collocates = Collocates(self.corpus, df_loc, p_query=p_query)
            tables[s] = collocates.show(
                window=window, order=order, cut_off=cut_off,
                ams=ams, min_freq=min_freq, frequencies=frequencies,
                flags=flags, marginals=marginals_loc
            )

        return tables



import os
import pytest
from pandas import read_csv


LOCAL = True
DIR_PATH = os.path.dirname(os.path.realpath(__file__))
DATA_PATH = os.path.join(DIR_PATH, 'data-dir')


@pytest.fixture
def germaparl():
    """ settings for small germaparl testing corpus """

    registry_path = os.path.join(DIR_PATH, "corpora/registry/")
    dump_path = os.path.join(DIR_PATH, "gold", "germaparl-seehofer.tsv")
    corpus_name = 'GERMAPARL1386'

    context = 50
    s_context = 'text'
    s_meta = 'text_id'

    s_query = 's'
    query = '[lemma="Seehofer"]'
    query_anchor = (
        r'@0[lemma="Horst"]? @1[lemma="Seehofer"] @2:[::]'
    )
    anchors = [0, 1, 2]
    query_within = query_anchor + " within " + s_query + ";"
    seehofer_dump = read_csv(dump_path, sep="\t", header=None, dtype=int,
                             index_col=[0, 1], names=['match', 'matchend'])

    return {
        'registry_path': registry_path,
        'corpus_name': corpus_name,
        'context': context,
        's_context': s_context,
        's_query': s_query,
        's_meta': s_meta,
        'query': query,
        'query_anchor': query_anchor,
        'anchors': anchors,
        'query_full': query_within,
        'dump': seehofer_dump
    }


@pytest.fixture
def brexit():
    """ settings for brexit tweet corpus """

    registry_path = (
        "/home/ausgerechnet/corpora/cwb/registry/"
    )
    corpus_name = "BREXIT_V20190522_DEDUP"
    lib_path = (
        "/home/ausgerechnet/implementation/spheroscope/library/BREXIT_V20190522_DEDUP/"
    )
    meta_path = (
        "/home/ausgerechnet/corpora/cwb/upload/"
        "brexit/brexit-preref-rant/brexit_v20190522_dedup.tsv.gz"
    )

    query_path = (os.path.join(DIR_PATH, "gold", "as_a_x_i_y_knowledge.json"))

    context = 50
    s_context = 'tweet'

    s_query = 'tweet'
    query = '[lemma="test"]'
    query_wordlist = (
        '[lemma = $verbs_cause]'
    )

    # according to experts ...
    query_argmin = {
        'cqp': (
            '("according" "to" | "as" [lemma = $verbs_communication] "by")'
            '<np> @0:[::][word != "(http.|www|t\\.co).+"]*'
            '(/name_any[] | [lemma = $nouns_experts |'
            'lemma = $nouns_person_profession]) [word != "(http.|www|t\\.co).+"]* </np>'
            '@1:[::] ","? <np>@2:[::][word != "(http.|www|t\\.co).+"]*</np>'
            '(<vp>[]*</vp>)+ (<np>[word != "(http.|www|t\\.co).+"]*</np>'
            ' | <vp>[]*</vp> |<pp>[]*</pp> | [pos_simple = "I|R"])+ @3:[::]'
        ),
        'corrections': {
            0: 0,
            1: -1,
            2: 0,
            3: -1
        },
        'slots': {
            'slot_single': 0,
            'slot_1': [0, 1],
            'slot_2': [2, 3]
        },
        'match_strategy': 'longest',
        's_query': 'tweet',
        's_context': 'tweet',
        's_show': ['tweet_id'],
        'p_text': 'word',
        'p_slots': 'lemma',
        'p_show': ['word', 'pos_ner', 'lemma', 'pos_ark']
    }

    return {
        'registry_path': registry_path,
        'corpus_name': corpus_name,
        'lib_path': lib_path,
        'meta_path': meta_path,
        'context': context,
        's_context': s_context,
        's_query': s_query,
        'query': query,
        'query_lib': query_wordlist,
        'query_argmin': query_argmin,
        'query_path': query_path
    }



from ccc.cache import Cache
from pandas import DataFrame
from random import randint
import os

from .conftest import DATA_PATH


def test_generate_idx():

    parameters = {'query': "test", 's_break': "test"}
    cache = Cache()
    s = cache.generate_idx(parameters)
    assert(isinstance(s, str))


def test_set_get():

    cache = Cache(os.path.join(DATA_PATH, 'test-cache'))
    parameters = {'query': "test", 's_break': "test", 'key': randint(0, 9e+9)}
    dump = DataFrame()
    r = cache.get(parameters.values())
    assert(r is None)
    cache.set(parameters, dump)
    r = cache.get(parameters)
    assert(r.empty)

    cache.set('testtest', dump)
    r = cache.get('testtest')
    assert(r.empty)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import shelve
from hashlib import sha256
# logging
import logging
logger = logging.getLogger(__name__)


class Cache:

    def __init__(self, path=None):
        self.path = path

    def generate_idx(self, identifiers, prefix='', length=10):
        string = ''.join([str(idx) for idx in identifiers])
        h = sha256(str(string).encode()).hexdigest()
        return prefix + h[:length]

    def get(self, identifier):

        if self.path is None:
            return None

        if isinstance(identifier, str):
            key = identifier
        else:
            key = self.generate_idx(identifier)

        with shelve.open(self.path) as db:
            if key in db.keys():
                logger.info('cache: retrieving object "%s"' % key)
                return db[key]
            else:
                return None

    def set(self, identifier, value):

        if self.path is None:
            return None

        if isinstance(identifier, str):
            key = identifier
        else:
            key = self.generate_idx(identifier)

        with shelve.open(self.path) as db:
            logger.info('cache: saving object "%s"' % key)
            db[key] = value



#! /usr/bin/env python
# -*- coding: utf-8 -*-


# Version 2.0 of "PyCQP" by Joerg Asmussen, DSL (Febr. 2008)
# Some changes by Philipp Heinrich (2020)


import sys
import os
import re
import random
import time
import subprocess
import select
import signal
from io import StringIO
from pandas import read_csv, DataFrame
from tempfile import NamedTemporaryFile
import threading
import logging
logger = logging.getLogger(__name__)


# GLOBAL CONSTANTS OF MODULE:
CPROGRESSCONTROLCYCLE = 5   # secs between each progress control cycle
CMAXREQUESTPROCTIME = 500   # max secs for processing a user request


# ERROR MESSAGE TYPES:
class ErrCQP:
    """Handle CQP error message."""
    def __init__(self, msg):
        """Class constructor."""
        self.msg = msg.rstrip()


class ErrKilled:
    """Handle errors when process is killed."""
    def __init__(self, msg):
        """Class constructor."""
        self.msg = msg.rstrip()


# ACTUAL INTERFACE:
class CQP:
    """Wrapper for CQP."""

    def _progressController(self):
        """Control the progress.

        CREATED: 2008-02

        This method is run as a thread.
        At certain intervals (CPROGRESSCONTROLCYCLE), it controls how long the
        CQP process of the current user has spent on processing this user's
        latest CQP command. If this time exceeds a certain maximun
        (CMAXREQUESTPROCTIME), this method kills the CQP process.
        """
        self.runController = True
        while self.runController:
            time.sleep(CPROGRESSCONTROLCYCLE)
            if self.execStart is not None:
                if time.time() - self.execStart > CMAXREQUESTPROCTIME *\
                 self.maxProcCycles:
                    logger.error(
                        "progress controller identified blocking cqp process id "
                        "{}".format(self.CQP_process.pid)
                    )
                    # os.kill(self.CQP_process.pid, SIGKILL) - doesn't work!
                    os.popen("kill -9 " + str(self.CQP_process.pid))  # works!
                    logger.error("=> killed!")
                    self.CQPrunning = False
                    break

    def __init__(self, binary="/usr/local/bin/cqp", options='-c', print_version=False):
        """Class constructor."""
        self.execStart = time.time()
        self.maxProcCycles = 1.0

        # start CQP as a child process of this wrapper
        if binary is None:
            logger.error("path to CQP binaries undefined")
            sys.exit(1)
        self.CQP_process = subprocess.Popen(binary + ' ' + options,
                                            shell=True,
                                            stdin=subprocess.PIPE,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE,
                                            universal_newlines=True,
                                            close_fds=True,
                                            preexec_fn=os.setsid)
        self.CQPrunning = True

        # init progress controller
        progressthread = threading.Thread(target=self._progressController, args=())
        progressthread.setDaemon(True)
        progressthread.start()

        # "cqp -c" should print version on startup:
        version_string = self.CQP_process.stdout.readline()
        version_string = version_string.rstrip()  # Equivalent to Perl's chomp
        self.CQP_process.stdout.flush()
        if print_version:
            logger.info(version_string)
        version_regexp = re.compile(
            r'^CQP\s+(?:\w+\s+)*([0-9]+)\.([0-9]+)(?:\.b?([0-9]+))?(?:\s+(.*))?$'
        )
        match = version_regexp.match(version_string)
        if not match:
            logger.error("CQP backend startup failed")
            sys.exit(1)
        self.major_version = int(match.group(1))
        self.minor_version = int(match.group(2))
        self.beta_version = int(match.group(3))
        self.compile_date = match.group(4)

        # We need cqp-2.2.b41 or newer (for query lock):
        if not (self.major_version >= 3 or (self.major_version == 2 and
                                            self.minor_version == 2 and
                                            self.beta_version >= 41)):
            logger.error("CQP version too old: " + version_string)
            sys.exit(1)

        # Error handling:
        self.error_handler = None
        self.status = 'ok'
        self.error_message = ''  # we store compound error messages as a STRING
        self.errpipe = self.CQP_process.stderr.fileno()

        # Debugging (prints more or less everything on stdout)
        self.debug = False

        # CQP defaults:
        self.Exec('set PrettyPrint off')
        self.execStart = None

    def Terminate(self):
        """Terminate controller thread, must be called before deleting CQP."""
        self.execStart = None
        self.runController = False

    def SetProcCycles(self, procCycles):
        """Set procCycles."""
        print("    Setting procCycles to {}".format(procCycles))
        self.maxProcCycles = procCycles
        return int(self.maxProcCycles * CMAXREQUESTPROCTIME)

    def __del__(self):
        """Stop running CQP instance."""
        if self.CQPrunning:
            # print "Deleting CQP with pid", self.CQP_process.pid, "...",
            self.CQPrunning = False
            self.execStart = time.time()
            if self.debug:
                logger.info("Shutting down CQP backend ...")
            self.CQP_process.stdin.write('exit;')  # exits CQP backend
            if self.debug:
                logger.info("Done\nCQP object deleted.")
            self.execStart = None
            # print "Finished"

    def __kill__(self):
        """ like self.__del__, but correct """
        self.Terminate()
        os.killpg(os.getpgid(self.CQP_process.pid), signal.SIGTERM)
        self.__del__()

    def Exec(self, cmd):
        """Execute CQP command.

        The method takes as input a command string and sends it
        to the CQP child process
        """
        self.execStart = time.time()
        self.status = 'ok'
        cmd = cmd.rstrip()  # Equivalent to Perl's 'chomp'
        cmd = re.sub(r';\s*$', r'', cmd)
        if self.debug:
            logger.info("CQP <<", cmd + ";")
        try:
            self.CQP_process.stdin.write(cmd + '; .EOL.;\n')
        except IOError:
            return None
        # In CQP.pm lines are appended to a list @result.
        # This implementation prefers a string structure instead
        # because output from this module is meant to be transferred
        # accross a server connection. To enable the development of
        # client modules written in any language, the server only emits
        # strings which then are to be structured by the client module.
        # The server does not emit pickled data according to some
        # language dependent protocol.
        self.CQP_process.stdin.flush()
        result = []
        while self.CQPrunning:
            ln = self.CQP_process.stdout.readline()
            ln = ln.strip()  # strip off whitespace from start and end of line
            if re.match(r'-::-EOL-::-', ln):
                if self.debug:
                    print("CQP " + "-" * 60)
                break
            if self.debug:
                print("CQP >> " + ln)
            if ln != '':
                result.append(ln)
            self.CQP_process.stdout.flush()
        self.Checkerr()
        self.execStart = None
        result = '\n'.join(result)
        result = result.rstrip()  # strip off whitespace from EOL (\n)
        return result

    def Query(self, query):
        """Execute query in safe mode (query lock)."""
        result = []
        key = str(random.randint(1, 1000000))
        errormsg = ''  # collect CQP error messages AS STRING
        ok = True      # check if any error occurs
        self.Exec('set QueryLock ' + key)  # enter query lock mode
        if self.status != 'ok':
            errormsg = errormsg + self.error_message
            ok = False
        result = self.Exec(query)
        if self.status != 'ok':
            errormsg = errormsg + self.error_message.decode('utf-8')
            ok = ok and False
        self.Exec('unlock ' + key)  # unlock with random key
        if self.status != 'ok':
            errormsg = errormsg + self.error_message
            ok = ok and False
        # Set error status & error message:
        if ok:
            self.status = 'ok'
        else:
            self.status = 'error'
            self.error_message = errormsg
        return result

    def Dump(self, subcorpus='Last', first=None, last=None):
        """Dump named query result into table of corpus positions."""

        # check first and last
        if first is None and last is None:
            # actual dump
            result = self.Exec('dump ' + subcorpus + ";")

        elif ((not isinstance(first, int) and first is not None) or
              (not isinstance(last, int) and last is not None)):
            logger.error("Invalid value for first (" + str(first) +
                         ") or last (" + str(last) + ") line in Dump() method")
            sys.exit(1)
        elif isinstance(first, int) and isinstance(last, int):
            if first > last:
                logger.error("Invalid value for first line (first = " +
                             str(first) + " > last = " + str(last) +
                             ") in Dump() method")
                sys.exit(1)
        else:
            if first is not None and last is None:
                last = first
            elif last is not None and first is None:
                first = last
            # actual dump
            result = self.Exec(
                'dump ' + subcorpus + " " + str(first) + " " + str(last) + ";"
            )

        # convert to pandas dataframe
        df = read_csv(StringIO(result),
                      sep="\t", header=None, index_col=[0, 1],
                      names=["match", "matchend", "target", "keyword"])
        df = df.astype(int)

        return df

    def Undump(self, subcorpus="Last", df=DataFrame()):
        """Undump named query result from table of corpus positions."""
        columns = []
        wth = ''
        if 'target' in df.columns:
            wth = 'with target '
            columns = ['target']
            if 'keyword' in df.columns:
                wth = 'with target keyword '
                columns = ['target', 'keyword']
        with NamedTemporaryFile(mode='wt') as f:
            f.write(str(len(df)) + "\n")
            df.to_csv(f, mode="a", sep="\t", columns=columns, header=None)
            f.seek(0)
            f.flush()
            self.Exec("undump " + subcorpus + " " + wth + '< "' + f.name + '";')

    def Group(self, subcorpus='Last',
              spec1='match.word', spec2='', cutoff='1'):
        """Compute frequency distribution over attribute values.

        (single values or pairs) using group command.

        Note that the arguments are specified in the logical order,
        in contrast to "group"
        """
        spec2_regexp = re.compile(r'^[0-9]+$')
        if spec2_regexp.match(spec2):
            cutoff = spec2
            spec2 = ''
        spec_regexp = re.compile(
          r'^(match|matchend|target[0-9]?|keyword)\.([A-Za-z0-9_-]+)$'
        )
        match = re.match(spec_regexp, spec1)
        if not match:
            logger.error("Invalid key '" + spec1 + "' in Group() method")
            sys.exit(1)
        spec1 = match.group(1) + ' ' + match.group(2)
        if spec2 != '':
            match = re.match(spec_regexp, spec2)
            if not match:
                logger.error("Invalid key '" + spec2 + "' in Group() method")
                sys.exit(1)
            spec2 = match.group(1) + ' ' + match.group(2)
            cmd = 'group ' + subcorpus + ' ' + spec2 + ' by ' + spec1 + \
                  ' cut ' + cutoff
        else:
            cmd = 'group ' + subcorpus + ' ' + spec1 + ' cut ' + cutoff
        result = self.Exec(cmd)
        return result

    def Count(self, subcorpus='Last', sort_clause=None, cutoff=1):
        """Compute frequency distribution for match strings.

        Based on sort clause.
        """
        if sort_clause is None:
            logger.error("Parameter 'sort_clause' undefined in Count() method")
            sys.exit(1)
        return self.Exec('count ' + subcorpus +
                         ' by ' + sort_clause +
                         ' cut ' + str(cutoff))

    def Checkerr(self):
        """Check CQP's stderr stream for error messages.

        (returns true if there was an error).
        OBS! In CQP.pm the error_message is stored in a list,
        In PyCQP_interface.pm we use a string (which better can be sent
        accross the server line).
        """
        ready = select.select([self.errpipe], [], [], 0)
        if self.errpipe in ready[0]:
            # We've got something on stderr -> an error must have occurred:
            self.status = 'error'
            self.error_message = self.Readerr()
        return not self.Ok()

    def Readerr(self):
        """Read all available lines from CQP's stderr stream."""
        return os.read(self.errpipe, 16384)

    def Status(self):
        """Read the CQP object's (error) status."""
        return self.status

    def Ok(self):
        """Simplified interface for checking for CQP errors."""
        if self.CQPrunning:
            return (self.Status() == 'ok')
        else:
            return False

    def Error_message(self):
        """Return the CQP error message."""
        if self.CQPrunning:
            return ErrCQP(self.error_message)
        else:
            msgKilled = '**** CQP KILLED ***\n\
            CQP COULD NOT PROCESS YOUR REQUEST\n'
            return ErrKilled(msgKilled + self.error_message)

    def Error(self, msg):
        """Processe/output error messages.

        (optionally run through user-defined error handler)
        """
        if self.error_handler is not None:
            self.error_handler(msg)
        else:
            logger.error(msg)

    def Set_error_handler(self, handler=None):
        """Set user-defined error handler."""
        self.error_handler = handler

    def Debug(self, on=False):
        """Switch debugging output on/off."""
        prev = self.debug
        self.debug = on
        return prev

    #########################
    # SOME ALIASES FOR NQRs #
    #########################
    def nqr_from_query(self, query, name='Last',
                       match_strategy='longest', return_dump=True):
        """Defines NQR from query, optionally returns dump.

        :param str query: valid CQP query
        :param str name: name for NQR
        :param str match_strategy: CQP matching strategy
        :param bool return_dump: whether to return the dump

        :return: df_dump
        :rtype: DataFrame

        """
        logger.info('defining NQR "%s" from query' % name)
        self.Query('%s=%s;' % (name, query))
        if return_dump:
            logger.info('dumping result')
            df_dump = self.Dump(name)
            return df_dump

    def nqr_from_dump(self, df_dump, name='Last'):
        """Alias for Undump. Defines NQR from given dump.

        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
                                  with optional columns 'target' and 'keyword'
        :param str name: name for NQR

        """
        logger.info('defining NQR "%s" from dump with %d matches' % (name, len(df_dump)))
        self.Undump(name, df_dump)

    def nqr_activate(self, corpus_name, name=None):
        """Activate NQR or whole corpus.

        :param str corpus_name: name of the corpus
        :param str name: name of NQR

        """
        if name is not None:
            logger.info('activating NQR "%s:%s"' % (corpus_name, name))
            self.Exec(name)
        else:
            logger.info('activating corpus "%s"' % corpus_name)
            self.Exec(corpus_name)

    def nqr_save(self, corpus_name, name='Last'):
        """Save NQR to disk.

        :param str corpus_name: name of the corpus
        :param str name: nqr to save

        """
        logger.info('saving NQR "%s:%s" to disk' % (corpus_name, name))
        self.Exec("save %s;" % name)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from random import sample
import itertools
# part of module
from .utils import node2cooc
# requirements
from pandas import DataFrame
# logging
import logging
logger = logging.getLogger(__name__)


class Concordance:
    """ concordancing """

    def __init__(self, corpus, df_dump):

        # bind corpus
        self.corpus = corpus

        # make sure there's context
        if 'context' not in df_dump.columns:
            df_dump['context'] = df_dump.index.get_level_values('match')
        if 'contextend' not in df_dump.columns:
            df_dump['contextend'] = df_dump.index.get_level_values('matchend')

        # bind dump and check size
        self.df_dump = df_dump
        self.size = len(df_dump)
        if self.size == 0:
            logger.warning('empty dump')

        # init anchor points
        anchors = [i for i in range(10) if i in df_dump.columns]
        self.anchors = anchors

    def simple(self, df, p_show=['word'], start='context', end='contextend'):
        """Retrieve concordance lines of provided df in 'simple'
        formatting, i.e. one column for each $p in p_show.

        """
        for p_att in p_show:
            df = self.corpus.dump2patt(
                df, p_att=p_att, start=start, end=end
            )

        return df[p_show]

    def kwic(self, df, p_show=['word']):
        """Retrieve concordance lines of provided df in 'kwic' formatting. The
        resulting dataframe has 3 times len(p_show) columns, namely
        for each $p in p_show:
        - left_$p: context .. match - 1
        - node_$p: match .. matchend
        - right_$p: matchend + 1 .. contextend

        """

        df = df.reset_index()
        df['leftend'] = df['match'] - 1
        df['rightstart'] = df['matchend'] + 1
        df = df.set_index(['match', 'matchend'])

        for p in p_show:
            df['left' + "_" + p] = self.corpus.dump2patt(
                df, p, start='context', end='leftend'
            )[p]
            df['node' + "_" + p] = self.corpus.dump2patt(
                df, p, start='match', end='matchend'
            )[p]
            df['right' + "_" + p] = self.corpus.dump2patt(
                df, p, start='rightstart', end='contextend'
            )[p]

        return df[["_".join(combo) for combo in itertools.product(
            ['left', 'node', 'right'], p_show
        )]]

    def slots(self, df, p_show=['word'], slots=None):
        """Retrieve concordance lines of provided df in 'slot' formatting.
        Slots are singletons or pairs of self.anchors.  By default,
        all single anchors are slots.

        If slots is a list, it will be translated into a dictionary
        with keys of the form "$slot[0]..$slot[1]".

        For each single slot $s and each p-attribute $p in p_show,
        $slot_$p is retrieved.  For pairs, $slot[0]..$slot[1]_$p is
        retrieved.

        If $slot[0] or $slot[1] is not defined in a row (hence the
        entry is -1), the pair is treated as a singleton.  If both
        positions are not defined, an empty string is returned.

        """

        if slots is None:
            slots = self.anchors + [('match', 'matchend')]

        # translate list into dict
        if isinstance(slots, list):
            slots_dict = dict()
            for slot in slots:
                if isinstance(slot, (int, str)):
                    key = slot
                else:
                    key = "..".join([str(s) for s in slot])
                slots_dict[key] = slot
            slots = slots_dict

        # init output with simple view
        df_lines = self.simple(df, p_show)

        # add slots
        for key in slots.keys():
            # allow definition of slots as singletons
            if isinstance(slots[key], (int, str)):
                start = end = slots[key]
            else:
                start, end = slots[key]
            df_lines[[str(key) + "_" + p for p in p_show]] = self.simple(
                df, p_show=p_show, start=start, end=end
            )[p_show]

        return df_lines

    def _export(self, index, row, p_show, form='dict'):
        """Translates one row of a df_dump into a dictionary or a DataFrame.

        :return: [cpos, offset, anchors] + p_show
        :rtype: dict or DataFrame

        """

        # pack all values into a dictionary
        row = dict(row)
        match, matchend = index
        row['match'] = match
        row['matchend'] = matchend

        # create cotext
        cotext = node2cooc(row)

        # lexicalize positions
        attribute_lists = zip(
            *list(map(lambda x: self.corpus.cpos2patts(x, p_show), cotext['cpos_list']))
        )

        # init output dictionary
        d = {
            'cpos': cotext['cpos_list'],
            'offset': cotext['offset_list']
        }
        # add attributes
        for p, a in zip(p_show, attribute_lists):
            d[p] = list(a)

        if form == 'dict':
            # add match as identifier
            d['match'] = match
            # process anchors
            d['anchors'] = dict()
            for a in self.anchors:
                d['anchors'][a] = row[a]
            return d

        elif form == 'dataframe':
            # convert to dataframe and sort columns
            df = DataFrame.from_records(d).set_index('cpos')
            df = df[['offset'] + p_show]
            # process anchors
            for a in self.anchors:
                df[a] = False
                if row[a] in df.index:
                    df.at[row[a], a] = True
            return df

    def dict(self, df, p_show=['word']):
        """Retrieve concordance lines of provided df in 'dict' formatting.

        Each dictionary has the following keys:
        - match: single integer as ID
        - cpos: list of corpus positions of tokens
        - offset: list of offset of token to match..matchend
        - $p for in p_show: list of tokens
        - anchors: dictionary of anchor: cpos

        All lists are aligned.

        """

        df['dict'] = df.apply(
            lambda row: self._export(row.name, row, p_show, form='dict'),
            axis=1
        )

        return df

    def dataframe(self, df, p_show=['word']):
        """Retrieve concordance lines of provided df_dict in 'dataframe'
        formatting.  df_dict must contain column 'dict' (return value
        of self.dict()).

        Each concordance line is a DataFrame with the cpos of each
        token as index:

        == (cpos), offset, p_show[0], p_show[1], ... ==

        """

        df['dataframe'] = df.apply(
            lambda row: self._export(row.name, row, p_show, form='dataframe'),
            axis=1
        )

        return df

    def lines(self, form='simple', p_show=['word'], s_show=[],
              order='first', cut_off=100, matches=None, slots=None):
        """Retrieve concordance lines from self.df_dump.  Central entry point
        for all methods.  Functionality includes:
        (1) simple / kwic / dict / slots / dataframe format
        (2) selection of p-attributes and s-attributes
        (3) selection of matches using cut-off and order or a
            pre-defined list of matches
        (4) definition of slots via anchor points (only for form='slot')

        Return value of all formats is a DataFrame indexed by (match,
        matchend) and requested s-attributes of the match:

        == (match, matchend), s_show[0], ... ==

        Further columns depend on output format, see the respective
        format implementations.

        :param str form: simple / kwic / dict / slots / dataframe
        :param list p_show: p-attributes to retrieve
        :param list s_show: s-attributes (at match position)
        :param str order: first / last / random
        :param int cut_off: how many lines to retrieve
        :param list matches: specific matches to retrieve
        :param dict slots: definition of slots via anchor points

        :return: concordance lines
        :rtype: DataFrame

        """

        # check parameters
        if len(self.df_dump) == 0:
            logger.error("no concordance lines to show")
            return
        if form not in ['simple', 'kwic', 'dict', 'dataframe', 'slots']:
            logger.error('format not implemented, using "simple" format')
            form = 'simple'
        if isinstance(p_show, str):
            p_show = [p_show]
        if isinstance(s_show, str):
            p_show = [s_show]

        # select appropriate subset of matches
        logger.info('lines: selecting matches')
        if matches is None:
            # select all matches
            matches = set(self.df_dump.index.droplevel('matchend'))
        # pre-process cut_off if necessary
        if (cut_off is None) or (len(matches) < cut_off):
            cut_off = len(matches)
        # order
        if order == 'random':
            matches = sample(matches, cut_off)
        elif order == 'first':
            matches = sorted(list(matches))[:cut_off]
        elif order == 'last':
            matches = sorted(list(matches))[-cut_off:]
        else:
            logger.error('order not implemented, using "random" order')
            order = 'random'
        logger.info("lines: retrieving %d concordance line(s)" % len(matches))
        df = self.df_dump.loc[matches, :]

        # retrieve p-attributes in respective formatting
        if len(p_show) > 0:
            if form == 'simple':
                df = self.simple(df, p_show)
            elif form == 'kwic':
                df = self.kwic(df, p_show)
            elif form == 'slots':
                df = self.slots(df, p_show, slots)
            elif form == 'dict':
                df = self.dict(df, p_show)
            elif form == 'dataframe':
                df = self.dataframe(df, p_show)

        # s-attributes
        for s_att in s_show:
            tmp = self.corpus.dump2satt(df, s_att)
            if s_att in tmp.columns:  # annotation
                df[s_att] = tmp[s_att]
            else:               # ID if no annotation
                df[s_att] = tmp[s_att + "_cwbid"]

        return df



#! /usr/bin/env python
# -*- coding: utf-8 -*-

""" frames.py: classes for dataframe types

- FreqFrame
- DumpFrame
- CoocFrame
- CollFrame
- ConcFrame

"""

import pandas as pd


@pd.api.extensions.register_dataframe_accessor("freq")
class FreqFrame:
    """A DataFrame for Frequencies:
    == (p_atts) freq ==
    MultiIndex, even if single p-att

    examples:
    == (word, ) freq ==
    == (lemma, pos) freq ==
    == (query, ) freq ==

    for MWUs, each p-att-layer is " ".joined
    """

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        # verify there is a column "freq" and a named MultiIndex
        if "freq" not in obj.columns or not isinstance(obj.index, pd.MultiIndex):
            raise AttributeError("wrong format")

    def test(self):
        print(self)



from ccc import Corpus
from ccc.collocates import Collocates
from ccc.keywords import Keywords

from .conftest import LOCAL, DATA_PATH

import pandas as pd
import pytest


def get_corpus(corpus_settings, data_path=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_path=corpus_settings['registry_path'],
        lib_path=corpus_settings.get('lib_path', None),
        data_path=data_path
    )


@pytest.mark.default
def test_collo_single(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\("] [lemma=".*"]+ [word="\\)"]'
    )
    df_dump = corpus.query(query).df
    collocates = Collocates(corpus, df_dump, 'lemma')
    c = collocates.show(order='log_likelihood')
    print(c)
    assert(type(c) == pd.DataFrame)
    assert('Dr.' in c.index)


@pytest.mark.default
def test_collo_combo(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\("] [lemma=".*"]+ [word="\\)"]'
    )
    df_dump = corpus.query(query).df
    collocates = Collocates(corpus, df_dump, ['lemma', 'pos'])
    c = collocates.show(order='log_likelihood')
    # print(c)
    assert(type(c) == pd.DataFrame)
    # assert('Dr.' in c.index)


@pytest.mark.fallback
def test_query_logging(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[word="\\("] [lemma=".*"]+ [word="\\)"]'
    )
    df_dump = corpus.query(query).df
    collocates = Collocates(corpus, df_dump, 'fail')
    c = collocates.show(order='log_likelihood', window=15)
    assert(type(c) == pd.DataFrame)
    assert('Dr.' in c.index)


def compare_counts(lemma, window, min_freq=0):
    # TODO: update to reproduceable example

    # CCC
    corpus = Corpus("GERMAPARL_1114")
    query = '[lemma="' + lemma + '"]'
    df_dump = corpus.query(query, context=window, context_break='s').df
    collocates = Collocates(corpus, df_dump, p_query='lemma')
    col = collocates.show(window=5, cut_off=None,
                          min_freq=min_freq)

    # UCS
    ucs = pd.read_csv("tests/gold/ucs-germaparl1114-" + lemma +
                      ".ds.gz", sep="\t", index_col=2, comment="#",
                      quoting=3, na_filter=False)
    ucs.index.name = 'item'
    try:
        O11_ucs_node = ucs.loc[lemma]['f']
        ucs.drop(lemma, inplace=True)
    except KeyError:
        O11_ucs_node = 0

    # identities that should hold between counting strategies
    # (1) N_ccc + f1_ccc = N_ucs
    # (2) f1_infl_ccc = f1_infl_ucs - O11_ucs_node
    nr = {
        'f1_ccc': int(corpus.marginals([lemma], "lemma")[['freq']].values[0]),
        'N_ccc': int(col[['N']].values[0]),
        'f1_infl_ccc': int(col[['f1']].values[0]),
        'N_ucs': int(ucs[['N']].values[0]),
        'f1_infl_ucs': int(ucs[['f1']].values[0]),
        'O11_ucs_node': O11_ucs_node
    }

    # make dataframes comparable
    ucs = ucs[['f', 'f2']]
    ucs.columns = ['O11', 'f2']
    ucs.sort_values(by=['O11', 'item'], ascending=False, inplace=True)

    assert(nr['N_ccc'] + nr['f1_ccc'] == nr['N_ucs'])
    assert(nr['f1_infl_ccc'] == nr['f1_infl_ucs'] - nr['O11_ucs_node'])


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.germaparl1114
@pytest.mark.collocates_gold
def test_compare_counts():
    compare_counts('Atomkraft', 5)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.germaparl1114
@pytest.mark.collocates_gold
def test_compare_counts_2():
    compare_counts('Angela', 5)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.germaparl1114
@pytest.mark.collocates_speed
@pytest.mark.skip(reason='takes too long')
def test_compare_counts_3():
    compare_counts('und', 2)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.germaparl1114
@pytest.mark.collocates_speed
def test_collocates_speed_many():
    corpus = Corpus("GERMAPARL_1114")
    query = '[lemma="sagen"]'
    df_dump = corpus.query(query, context=2, context_break='s').df
    collocates = Collocates(corpus, df_dump, p_query='lemma')
    c2 = collocates.show(window=2, cut_off=50)
    assert(type(c2) == pd.DataFrame)


@pytest.mark.persistence
def test_collocates_persistence(germaparl):
    corpus = get_corpus(germaparl)
    query_1 = (
        '"SPD"'
    )
    query_2 = (
        '"CSU"'
    )

    # will show collocates for query_1
    result = corpus.query(query_1, context_break='s').df
    collocates = Collocates(corpus, result, 'lemma')
    line_1 = collocates.show()

    # will show collocates for query_1
    result = corpus.query(query_2, context_break='s').df
    line_2 = collocates.show()

    # will show collocates for query_2
    collocates = Collocates(corpus, result, 'lemma')
    line_3 = collocates.show()

    assert(line_1.equals(line_2))
    assert(not line_2.equals(line_3))


@pytest.mark.keywords_collocates
def test_query_keywords_collocates(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '"Horst" expand to s'
    )
    dump = corpus.query(query)
    keywords = Keywords(corpus, df_dump=dump.df, p_query='lemma')
    assert('Horst' == keywords.show(order='log_likelihood').head(1).index[0])


@pytest.mark.mwu_marginals
def test_collocates_mwu(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="CDU"] "/"? [lemma="CSU"]?'
    )
    result = corpus.query(query, match_strategy='longest')
    collocates = Collocates(corpus, result.df, 'lemma')
    c = collocates.show(order='log_likelihood', cut_off=None)
    assert(type(c) == pd.DataFrame)
    assert(len(c) > 9)
    assert('CSU' in c.index)
    assert(int(c.loc['CSU']['in_nodes']) > int(c.loc['CSU']['f']))


@pytest.mark.fold_items
def test_collocates_pp(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '"SPD"'
    )
    result = corpus.query(query)
    collocates = Collocates(corpus, result.df, p_query='word')
    c = collocates.show(order='log_likelihood', cut_off=None)
    assert(int(c.loc['Die']['O11']) < int(c.loc['die']['O11']))
    c = collocates.show(order='log_likelihood', cut_off=None, flags="%cd")
    assert('die' in c.index and 'Die' not in c.index)


@pytest.mark.fail
def test_collocates_empty(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Armin"]? [lemma="NAHH"]'
    )
    dump = corpus.query(query)
    collocates = Collocates(corpus, dump.df, p_query='word')
    collocates.show()


@pytest.mark.fail
def test_collocates_no_context(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Armin"]? [lemma="Laschet"]'
    )
    dump = corpus.query(query)
    collocates = Collocates(corpus, dump.df, p_query='word', mws=0)
    collocates.show()


@pytest.mark.fail
def test_collocates_no_mws(germaparl):
    corpus = get_corpus(germaparl)
    query = (
        '[lemma="Armin"]? [lemma="Laschet"]'
    )
    dump = corpus.query(query)
    collocates = Collocates(corpus, dump.df, p_query='word', mws=None)
    collocates.show()


def test_collocates_nodes(germaparl):

    corpus = get_corpus(germaparl)

    query = (
        '[lemma=","] | [lemma="\\."] | [lemma="\\)"] | [lemma="\\("]'
    )
    # three discoursemes
    dump = corpus.query(query)
    print(dump.df)
    collocates = Collocates(corpus, dump.df)
    df = collocates.show(cut_off=None)
    assert("," not in df.index)
    assert("(" not in df.index)



from ccc.cqp import CQP
import pandas as pd


def test_cqp_version():
    print()
    print("... you should see your CQP version below ...")
    CQP(print_version=True)


def test_cqp_kill():

    from time import sleep
    n = 100
    rate = 1000                 # per second
    print()
    print("... spawning several cqp processes ...")
    print("... check htop or equivalent ...")
    for i in range(n):
        cqp = CQP()
        print("process_id = %d, run = %d/%d" % (
            cqp.CQP_process.pid,
            i + 1, n
            ), end="\r"
        )
        cqp.__kill__()
        sleep(1/rate)
    print()


def test_cqp_dump(germaparl):
    cqp = CQP(
        binary="cqp",
        options='-c -r ' + germaparl['registry_path']
    )
    cqp.Exec(germaparl['corpus_name'])
    cqp.Query('"Horst"')
    df = cqp.Dump()
    assert(len(df) == 55)
    assert(isinstance(df, pd.DataFrame))


def test_cqp_undump(germaparl):
    cqp = CQP(
        binary="cqp",
        options='-c -r ' + germaparl['registry_path']
    )
    cqp.Exec(germaparl['corpus_name'])
    cqp.Query('"Horst"')
    df = cqp.Dump()
    cqp.Undump("Test", df)
    assert(int(cqp.Exec("size Test;")) > 0)
    assert(cqp.Exec("size Test;") == str(len(df)))


def test_cqp_group(germaparl):
    cqp = CQP(
        binary="cqp",
        options='-c -r ' + germaparl['registry_path']
    )
    cqp.Exec(germaparl['corpus_name'])
    cqp.Query('[lemma="Horst"] [lemma="Seehofer"]')
    counts = cqp.Group(spec1="match.lemma", spec2="matchend.lemma")
    assert(type(counts) == str)
    assert(int(counts.split("\t")[-1]) == 11)


def test_nqr_from_query(germaparl):
    cqp = CQP(
        binary="cqp",
        options='-c -r ' + germaparl['registry_path']
    )
    cqp.Exec(germaparl['corpus_name'])
    assert(int(cqp.Exec('size Seehofer;')) == 0)
    cqp.nqr_from_query(
        query='[lemma="Seehofer"];',
        name='Seehofer',
        return_dump=False
    )
    assert(int(cqp.Exec('size Seehofer;')) > 0)
    cqp.__kill__()


def test_nqr_from_dump(germaparl):
    cqp = CQP(
        binary="cqp",
        options='-c -r ' + germaparl['registry_path']
    )
    cqp.Exec(germaparl['corpus_name'])
    assert(int(cqp.Exec('size Seehof;')) == 0)
    df_dump = germaparl['dump']
    cqp.nqr_from_dump(df_dump, name='Seehof')
    assert(int(cqp.Exec('size Seehof;')) > 0)
    cqp.__kill__()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
from io import StringIO
from glob import glob
# part of module
from .cqp import CQP
from .cache import Cache
from .counts import Counts
from .utils import preprocess_query
from .utils import chunk_anchors, correct_anchors
from .dumps import Dump
from .counts import cwb_scan_corpus
# requirements
from CWB.CL import Corpus as Attributes
from pandas import DataFrame, read_csv
from pandas.errors import EmptyDataError
from numpy import minimum, maximum
# logging
import logging
logger = logging.getLogger(__name__)


def start_cqp(cqp_bin, registry_path,
              data_path=None, corpus_name=None,
              lib_path=None, subcorpus=None):
    """Start CQP process, activate (sub-)corpus, set paths, and read
    library (macros and wordlists).

    :param str cqp_bin: /path/to/cqp-binary
    :param str registry_path: /path/to/cwb/registry/
    :param str data_path: /path/to/data/and/cache/
    :param str corpus_name: name of corpus in CWB registry
    :param str lib_path: /path/to/macros/and/wordlists/
    :param str subcorpus: name of subcorpus (NQR)

    :return: CQP process
    :rtype: CQP

    """

    cqp = CQP(
        binary=cqp_bin,
        options='-c -r ' + registry_path
    )

    if data_path is not None:
        cqp.Exec('set DataDirectory "%s"' % data_path)
    if corpus_name is not None:
        cqp.Exec(corpus_name)
    if subcorpus is not None:
        cqp.Exec(subcorpus)

    if lib_path is not None:

        # wordlists
        wordlists = glob(os.path.join(lib_path, 'wordlists', '*'))
        for wordlist in wordlists:
            name = wordlist.split('/')[-1].split('.')[0]
            abs_path = os.path.abspath(wordlist)
            cqp_exec = 'define $%s < "%s";' % (name, abs_path)
            cqp.Exec(cqp_exec)

        # macros
        macros = glob(os.path.join(lib_path, 'macros', '*'))
        for macro in macros:
            abs_path = os.path.abspath(macro)
            cqp_exec = 'define macro < "%s";' % abs_path
            cqp.Exec(cqp_exec)
        # execute each macro once (avoids CQP shortcoming for nested macros)
        macros = cqp.Exec("show macro;").split("\n")
        for macro in macros:
            cqp.Exec(macro)

    return cqp


class Corpora:
    """Interface to CWB-indexed corpora."""

    def __init__(self, cqp_bin='cqp',
                 registry_path='/usr/local/share/cwb/registry/'):
        """Establish connection to registry and CQP.

        :param str cqp_bin: /path/to/cqp-binary
        :param str registry_path: /path/to/cwb/registry/

        """

        self.registry_path = registry_path
        self.cqp_bin = cqp_bin

    def __str__(self):
        """Method for printing.

        :return: available corpora
        :rtype: str
        """

        return "\n" + "\n".join([
            'registry-path: "%s"' % self.registry_path,
            'cqp-bin      : "%s"' % self.cqp_bin,
            "corpora:",
            self.show().to_string(),
        ])

    def show(self):
        """Show all corpora defined in registry and available via CQP
        alongside corpus size (number of tokens).

        :return: available corpora
        :rtype: DataFrame

        """

        # get all corpora defined in registry
        cqp = start_cqp(self.cqp_bin, self.registry_path)
        corpora = cqp.Exec("show corpora;").split("\n")
        cqp.__kill__()

        # check availability and corpus sizes
        corpora_available = list()
        sizes = list()
        for corpus_name in corpora:

            try:
                corpus = Corpus(
                    corpus_name, cqp_bin=self.cqp_bin,
                    registry_path=self.registry_path, data_path=None
                )
                corpora_available.append(corpus_name)
                sizes.append(corpus.corpus_size)

            except SystemError:
                logger.warning(
                    'corpus "%s" defined in registry but not available' % corpus_name
                )

        # create dataframe
        corpora = DataFrame({'corpus': corpora_available,
                             'tokens': sizes})
        corpora = corpora.set_index('corpus')

        return corpora

    def activate(self, corpus_name,
                 lib_path=None, data_path='/tmp/ccc-data/'):

        return Corpus(corpus_name,
                      lib_path=lib_path,
                      cqp_bin=self.cqp_bin,
                      registry_path=self.registry_path,
                      data_path=data_path)


class Corpus:
    """Interface to CWB-indexed corpus.

    After initializing, the corpus class has ...

    ... the following attributes:
    .data_path
    .registry_path
    .cqp_bin
    .lib_path
    .corpus_name
    .subcorpus [None]
    .attributes_available
    .corpus_size

    ... the following initialized classes:
    .attributes
    .cache
    .counts

    ... the following methods:
    .__str__
    ._attributes_available
    .start_cqp
    .copy
    .cpos2patts                 # p-attributes
    .marginals                  # p-attributes
    .cpos2sid                   # s-attributes
    .show_nqr                   # subcorpora
    .dump_from_s_att            # creating dumps
    .dump_from_query            # creating dumps
    .dump2patt                  # working on dumps
    .dump2satt                  # working on dumps
    .dump2context               # working on dumps
    .query_s_att                # query alias
    .query                      # query alias

    """

    def __init__(self, corpus_name, lib_path=None, cqp_bin='cqp',
                 registry_path='/usr/local/share/cwb/registry/',
                 data_path='/tmp/ccc-data/'):
        """Establish connection to CQP and corpus attributes, set paths, read
        library. Raises KeyError if corpus not in registry.

        :param str corpus_name: name of corpus in CWB registry
        :param str lib_path: /path/to/macros/and/wordlists/
        :param str cqp_bin: /path/to/cqp-binary
        :param str registry_path: /path/to/cwb/registry/
        :param str data_path: /path/to/data/and/cache/

        """

        # process data path
        if data_path is not None:
            if not data_path.endswith(corpus_name):
                data_path = os.path.join(data_path, corpus_name)
            self.data_path = data_path
            if not os.path.isdir(self.data_path):
                os.makedirs(self.data_path)
            cache_path = os.path.join(self.data_path, "CACHE")
        else:
            self.data_path = None
            cache_path = None

        # set registry path and cqp_bin
        self.registry_path = registry_path
        self.cqp_bin = cqp_bin

        # macros and wordlists
        self.lib_path = lib_path

        # init (sub-)corpus information
        self.corpus_name = corpus_name
        self.subcorpus = None

        # init attributes
        self.attributes = Attributes(self.corpus_name, registry_dir=self.registry_path)

        # get corpus size
        self.corpus_size = len(self.attributes.attribute('word', 'p'))

        # get available corpus attributes
        self.attributes_available = self._attributes_available()

        # init cache
        self.cache = Cache(cache_path)

        # init counts
        self.counts = Counts(self.corpus_name, self.registry_path)

    def __str__(self):
        """Method for printing.

        :return: corpus_name, corpus_size, data_path, subcorpus
        :rtype: str

        """

        return "\n".join([
            'a ccc.Corpus: "%s"' % self.corpus_name,
            "size        : %s" % str(self.corpus_size),
            "data        : %s" % str(self.data_path),
            "subcorpus   : %s" % str(self.subcorpus),
            "attributes  :",
            self.attributes_available.to_string(),
        ])

    def _attributes_available(self):
        """Get indexed p- and s-attributes. Will be run once when initializing
        the corpus.

        :return: attributes and annotation info
        :rtype: DataFrame

        """

        # use CQP's context descriptor
        cqp = self.start_cqp()
        cqp_ret = cqp.Exec('show cd;')
        cqp.__kill__()

        # read as dataframe
        attributes = read_csv(
            StringIO(cqp_ret), sep='\t',
            names=['type', 'attribute', 'annotation', 'active']
        ).fillna(False)

        # post-process Boolean columns
        attributes['active'] = (attributes['active'] == "*")
        attributes['annotation'] = (attributes['annotation'] == '-V')

        return attributes

    def start_cqp(self):
        """Start CQP process.

        :return: CQP process
        :rtype: CQP

        """

        return start_cqp(
            self.cqp_bin,
            self.registry_path,
            self.data_path,
            self.corpus_name,
            self.lib_path,
            self.subcorpus
        )

    def copy(self):
        """Get a fresh initialization of the corpus.

        :return: corpus
        :rtype: Corpus

        """

        return Corpus(
            self.corpus_name,
            self.lib_path,
            self.cqp_bin,
            self.registry_path,
            self.data_path
        )

    ################
    # p-attributes #
    ################
    def cpos2patts(self, cpos, p_atts=['word'], ignore=True):
        """Retrieve p-attributes of corpus position.

        :param int cpos: corpus position to fill
        :param list p_atts: p-attribute(s) to fill position with
        :param bool ignore: whether to return (None, .*) for -1

        :return: p-attribute(s) at cpos
        :rtype: tuple

        """

        if cpos == -1 and ignore:
            token = [None] * len(p_atts)
        else:
            token = [
                self.attributes.attribute(p_att, 'p')[cpos] for p_att in p_atts
            ]

        return tuple(token)

    def marginals(self, items, p_att='word', flags=0, pattern=False):
        """Extract marginal frequencies for given unigrams or unigram patterns
        of a single p-attribute.  0 if not in corpus.  For
        combinations of p-attributes, see marginals_complex.

        :param list items: items to get marginals for
        :param str p_att: p-attribute to get frequencies for
        :param int flags: 1 = %c, 2 = %d, 3 = %cd (will activate wildcards)
        :param bool pattern: activate wildcards?

        :return: frequencies of the items in the whole corpus indexed by items
        :rtype: FreqFrame

        """

        pattern = True if flags > 0 else pattern

        tokens_all = self.attributes.attribute(p_att, 'p')

        # loop through items and collect frequencies
        counts = list()
        for item in items:
            if not pattern:
                try:
                    counts.append(tokens_all.frequency(item))
                except KeyError:
                    counts.append(0)
            else:
                cpos = tokens_all.find_pattern(item, flags=flags)
                counts.append(len(cpos))

        # create dataframe
        df = DataFrame({'freq': counts, p_att: items})
        df = df.set_index(p_att)

        return df

    def marginals_complex(self, items, p_atts=['word']):
        """Extract marginal frequencies for p-attribute combinations,
        e.g. ["word", "lemma"].  0 if not in corpus.  Marginals are
        retrieved using cwb-scan-corpus, result is cached.

        :param list items: list of tuples
        :param list p_atts: list of p-attributes

        :return: counts of the items in the whole corpus indexed by items
        :rtype: FreqFrame

        """

        # retrieve all marginals for p-att combination from cache if possible
        identifier = "_".join(p_atts) + "_marginals"
        df = self.cache.get(identifier)
        if df is not None:
            logger.info('using cached version of marginals of "%s"' % "_".join(p_atts))
        else:
            # calculate all marginals for p-att combination
            df = cwb_scan_corpus(None, self.corpus_name, self.registry_path, p_atts)
            self.cache.set(identifier, df)

        # select relevant rows
        df = df.reindex(items)
        df = df.fillna(0, downcast='infer')
        return df

    ################
    # s-attributes #
    ################
    def cpos2sid(self, cpos, s_att):
        """Get cwb-id of s-att at cpos. -1 if not present.

        :param int cpos: corpus position
        :param str s_atts: s-attribute to get cwb-id for

        :return: cwb-id
        :rtype: int

        """
        s_attributes = self.attributes.attribute(s_att, "s")
        try:
            return s_attributes.cpos2struc(cpos)
        except KeyError:
            return -1

    ##############
    # subcorpora #
    ##############
    def show_nqr(self):
        """Get subcorpora defined in CQP as DataFrame.

        :return: available subcorpora
        :rtype: DataFrame

        """
        cqp = self.start_cqp()
        cqp_return = cqp.Exec("show named;")
        try:
            df = read_csv(StringIO(cqp_return), sep="\t", header=None)
            df.columns = ["storage", "corpus:subcorpus", "size"]
            crpssbcrps = df["corpus:subcorpus"].str.split(":", 1).str
            df['corpus'] = crpssbcrps[0]
            df['subcorpus'] = crpssbcrps[1]
            df.drop('corpus:subcorpus', axis=1, inplace=True)
            df = df[['corpus', 'subcorpus', 'size', 'storage']]
        except EmptyDataError:
            logger.info("no subcorpora defined")
            df = DataFrame()

        cqp.__kill__()
        return df

    def activate_subcorpus(self, nqr=None, df_dump=None):
        """Activate subcorpus.  If no df_dump is given, this sets
        self.subcorpus and logs an error if subcorpus not defined.  If
        a df_dump is given, the df_dump will be undumped.

        :param str subcorpus: subcorpus name defined in CQP
        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
                                  with optional columns 'target' and 'keyword'

        """

        if df_dump is not None:
            cqp = self.start_cqp()
            cqp.nqr_from_dump(df_dump, nqr)
            cqp.nqr_save(self.corpus_name, nqr)
            cqp.__kill__()

        if nqr is not None:
            # raise an error if subcorpus not available
            if nqr not in self.show_nqr()['subcorpus'].values:
                logger.error('subcorpus "%s" not defined)' % nqr)
                self.activate_subcorpus()
            else:
                logger.info('switched to subcorpus "%s"' % nqr)
        else:
            logger.info('switched to corpus "%s"' % self.corpus_name)

        # activate subcorpus
        self.subcorpus = nqr

    ##################
    # CREATING DUMPS #
    ##################
    def dump_from_s_att(self, s_att, annotation=True):
        """Create s-attribute spans as DataFrame of corpus positions.
        Resulting df_dump is indexed by (match, matchend).

        Note that s-attribute values are not indexed by the CWB.
        CWB.CL implementation just iterates over all annotations.

        This method thus creates a dataframe with the
        (non-overlapping) spans encoded as matches

        === (match, matchend) $s_cwbid, $s* ===

        and caches the result.

        $s is only created if annotation is True and attribute is
        actually annotated.

        :param str s_att: s-attribute to get spans and annotation for
        :param bool annotation: whether to retrieve annotation (if present)

        :return: df_dump
        :rtype: DataFrame

        """

        # retrieve from cache if possible
        identifier = s_att + "_spans"
        df = self.cache.get(identifier)
        if df is not None:
            logger.info('using cached version of spans of "%s"' % s_att)
            return df

        # compute
        logger.info('creating dataframe of spans of "%s"' % s_att)

        df = DataFrame(list(self.attributes.attribute(s_att, 's')))

        # two or three columns: 0 (start), 1 (end), 2* (annotation)
        if annotation:
            annotation = (2 in df.columns)  # just to make sure ...
            if not annotation:
                logger.info('s-att "%s" does not have any annotation' % s_att)
            else:
                df[2] = df[2].apply(lambda x: x.decode('utf-8'))

        # post-process
        df = df.reset_index()
        df = df.rename({'index': s_att + '_cwbid',
                        0: 'match',
                        1: 'matchend',
                        2: s_att}, axis=1)
        df = df.set_index(['match', 'matchend'])

        # put into cache
        self.cache.set(identifier, df)

        return df

    def dump_from_query(self, query, s_query=None, anchors=[],
                        match_strategy='standard', name='Last', save=False):
        """Execute query, get DataFrame of corpus positions (CWB dump).
        Resulting df_dump is indexed by (match, matchend).

        Note that in the CWB, only two anchors can be active
        simultaneously. The method thus runs the query once with
        anchors set to [0, 1], and then collects the remaining anchor
        points by running the query again on the NQR of the first
        query run for each pair of remaining anchor points.  Optional
        columns for each anchor:

        === (match, matchend), 0*, ..., 9* ===

        The result is cached.

        :param str query: valid CQP query (without 'within' clause)
        :param str s_query: s-attribute used for initial query
        :param list anchors: anchors to search for
        :param str match_strategy: CQP matching strategy
        :param str name: name for NQR
        :param bool save: whether to save NQR to disk

        :return: df_dump
        :rtype: DataFrame

        """

        # identify query
        if self.subcorpus is not None:
            # check subcorpus size to avoid confusion when re-naming
            cqp = self.start_cqp()
            sbcrpssize = cqp.Exec("size %s" % self.subcorpus)
            cqp.__kill__()
        else:
            sbcrpssize = None
        identifier = self.cache.generate_idx([
             query, s_query, anchors, match_strategy, self.subcorpus, sbcrpssize
        ], prefix="df_dump:")

        # retrieve from cache if possible
        df_dump = self.cache.get(identifier)
        if df_dump is not None:
            logger.info(
                'using cached version "%s" of df_dump with %d matches' % (
                    identifier, len(df_dump)
                )
            )
            return df_dump

        # init cqp and set matching strategy
        cqp = self.start_cqp()
        cqp.Exec('set MatchingStrategy "%s";' % match_strategy)

        # include optional within clause
        if s_query is None:
            start_query = query
        else:
            start_query = query + ' within ' + s_query

        # first run: anchors at 0 and 1 (considering within clause)
        logger.info("running CQP query")
        cqp.Exec('set ant 0; ank 1;')
        df_dump = cqp.nqr_from_query(
            query=start_query,
            name=name,
            match_strategy=match_strategy,
            return_dump=True
        )
        df_dump.columns = [0, 1]
        logger.info("found %d matches" % len(df_dump))

        # if there's nothing to return ...
        if df_dump.empty:
            cqp.__kill__()
            return df_dump

        # join all other anchors
        remaining_anchors = list(chunk_anchors(anchors, 2))
        if len(remaining_anchors) > 0:

            # restrict subsequent queries on initial matches
            cqp.nqr_activate(self.corpus_name, name)

            for pair in remaining_anchors:
                logger.info(".. running query for anchor(s) %s" % str(pair))
                # set appropriate anchors
                cqp.Exec('set ant %d;' % pair[0])
                if len(pair) == 2:
                    cqp.Exec('set ank %d;' % pair[1])
                else:
                    cqp.Exec('set ank %d;' % 1)
                # dump new anchors
                cqp.Query('tmp = <match> ( %s );' % query)
                df = cqp.Dump("tmp")
                # select columns and join to global df
                if len(pair) == 2:
                    df.columns = [pair[0], pair[1]]
                else:
                    df.columns = [pair[0], 1]
                    df = df.drop(1, axis=1)
                df_dump = df_dump.join(df)

            # NA handling
            logger.info("post-processing dataframe")
            df_dump = df_dump.dropna(axis=1, how='all')
            df_dump = df_dump.fillna(-1, downcast='integer')

        # restrict output to requested anchors
        df_dump = df_dump[anchors]

        # put into cache
        self.cache.set(identifier, df_dump)

        if save:
            cqp.nqr_save(self.corpus_name, name)

        cqp.__kill__()

        return df_dump

    #################################################
    # WORKING ON DUMPS ##############################
    #################################################
    def _dump2patt_row(self, row, p_att, start, end):
        """Retrieve p-attribute annotation from start to end of one row.

        :param Series row: dataframe row that contain start and end keys
        :param str p_att: p-attribute to retrieve
        :param str start: key of start column (int or str)
        :param str end: key of end column (int or str)

        :return: p-attribute annotation
        :rtype: str
        """

        cpos_start = row[start]
        cpos_end = row[end]

        # if both are missing, return empty string
        if cpos_start == cpos_end == -1:
            return ""
        # if one of them is missing, set start = end or end = start
        if cpos_start == -1:
            cpos_start = cpos_end
        if cpos_end == -1:
            cpos_end = cpos_start

        # lexicalize
        p = self.attributes.attribute(p_att, 'p')
        return " ".join(p[int(cpos_start): int(cpos_end) + 1])

    def dump2patt(self, df_dump, p_att='word', start='match', end='matchend'):
        """Retrieve p-attribute annotation from start to end.

        === (match, matchend), $p ===

        Any additional columns of df_dump are preserved as long as
        there are no conflicts (in which case the original column will
        be overwritten).

        :param DataFrame df_dump: DataFrame with specified columns (possibly as index)
        :param str p_att: p-attribute to retrieve
        :param str start: key of start column (int or str)
        :param str end: key of end column (int or str)

        :return: df_dump
        :rtype: DataFrame

        """

        # pre-process
        index_names = df_dump.index.names
        df = df_dump.reset_index()

        # retrieve attribute
        df[p_att] = df.apply(
            lambda row: self._dump2patt_row(row, p_att, start, end), axis=1
        )

        # post-process
        df = df.set_index(index_names)

        return df

    def dump2satt(self, df_dump, s_att, annotation=True):
        """Retrieve cwb-id, span, and annotation of s-attribute at match.

        Note that this only takes into account the match, not the
        matchend. This is reasonable assuming that the retrieved s-att
        comprises complete matches (match..matchend).

        === (match, matchend), $s_cwbid, $s_span, $s_spanend, $s*  ===

        $s is only created if annotation is True and attribute is
        actually annotated.

        Any additional columns of df_dump are preserved as long as
        there are no conflicts (in which case the original columns
        will be overwritten).

        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
        :param str s_att: s-attribute to retrieve
        :param bool annotation: whether to retrieve annotation of s-att

        :return: df_dump
        :rtype: DataFrame

        """

        # init dataframe
        df = df_dump.reset_index()[['match', 'matchend']]

        # get $s_id
        df[s_att + "_cwbid"] = df.match.apply(
            lambda x: self.cpos2sid(x, s_att)
        )
        nr_missing = (df[s_att + '_cwbid'] == -1).sum()
        logger.info('s-att "%s" exists at %d of %d matches' % (
            s_att, len(df) - nr_missing, len(df)
        ))

        # retrieve where possible
        s_attributes = self.attributes.attribute(s_att, "s")
        df_s = df.loc[~(df[s_att + "_cwbid"] == -1)]
        df = df.join(DataFrame(
            index=df_s.index,
            data=df_s[s_att + "_cwbid"].apply(lambda x: s_attributes[int(x)]).to_list()
        ))

        # two or three columns: 0 (start), 1 (end), 2* (annotation)
        if annotation:
            annotation = (2 in df.columns)  # just to make sure ...
            if not annotation:
                logger.info('s-att "%s" does not have any annotation' % s_att)
            else:
                df[2] = df[2].apply(lambda x: x.decode('utf-8'))

        # restore original index and post-process
        df = df.set_index(['match', 'matchend'])
        df = df.rename({0: s_att + '_span',
                        1: s_att + '_spanend',
                        2: s_att}, axis=1)

        # join to original dataframe
        df_dump = df_dump.join(df, lsuffix='_bak')
        df_dump = df_dump[[col for col in df_dump if not str(col).endswith('_bak')]]
        df_dump[[
            s_att + '_span', s_att + '_spanend'
        ]] = df[[
            s_att + '_span', s_att + '_spanend'
        ]].fillna(-1, downcast='infer')

        return df_dump

    def dump2context(self, df_dump, context_left, context_right, context_break):
        """Extend df_dump to context, breaking the context at context_break.

        === (match, matchend), contextid*, context, contextend  ===

        Columns for $context_break, $context_break_cwbid,
        $context_break_span, $context_break_spanend, are also created
        (if applicable).

        Any additional columns of df_dump are preserved.

        Note that in contrast to matches, contexts may overlap.

        For positions where the s-att specified by context_break does
        not exist, context_break is ignored.

        The context creation algorithm does not take into account that
        match and matchend may be part of different spans defined by
        context_break; it only looks at the s-attributes of match, not
        of matchend.

        For the _context_ column (left hand side), the strategy is as
        follows; the strategy for _contextend_ (right hand side) is
        analogous (using context_right, matchend and
        context_break_spanend).
        if context_break_span is None and context_left is None
            => context = match
        if context_break_span is None and context_left is not None
            => context = max(0, match - context_left)
        if context_break_span is not None and context_left is None
            => context = context_break_span
        if context_break_span is not None and context_left is not None
            => context = max(match - context_left, s_start)

        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
        :param int context_left: maximum context to the left of match
        :param int context_right: maximum context to the right of matchend
        :param str context_break: s-attribute to confine context

        """

        if context_break is None:
            df = df_dump.reset_index()
            # left
            if context_left is None:
                df['context'] = df.match
            else:
                df['context'] = maximum(0, df.match - context_left)
            # right
            if context_right is None:
                df['contextend'] = df.matchend
            else:
                df['contextend'] = minimum(
                    self.corpus_size - 1, df.matchend + context_right
                )
        else:
            # get context confined by s-attribute
            df = self.dump2satt(df_dump, context_break, annotation=False).reset_index()

            # save contextid
            df['contextid'] = df[context_break + '_cwbid']

            # replace -1 to not confuse min()
            df[[
                context_break + '_spanend'
            ]] = df[[context_break + '_spanend']].replace(-1, self.corpus_size + 1)

            # left
            if context_left is None:
                df['context'] = df[context_break + '_span']
            else:
                df['context'] = df.match - context_left
                df['context'] = df[[
                    'context', context_break + '_span'
                ]].max(axis=1)
            # right
            if context_right is None:
                df['contextend'] = df[context_break + '_spanend']
            else:
                df['contextend'] = df.matchend + context_right
                df['contextend'] = df[[
                    'contextend', context_break + '_spanend'
                ]].min(axis=1)

            # revert replacement
            df[[
                context_break + '_spanend'
            ]] = df[[context_break + '_spanend']].replace(self.corpus_size + 1, -1)

        # restore original index
        df = df.set_index(['match', 'matchend'])

        return df

    #################################################
    # QUERY ALIASES #################################
    #################################################
    def query_s_att(self, s_att, values=set(), name=None):
        """Get s-attribute spans as Dump, optionally restricting the spans by
        matching the provided values against the s-att annotations.

        === (match, matchend), $s_cwbid, $s* ===

        :param str s_att: s-attribute to use for spans
        :param set values: values of s-att annotation to restrict spans to

        :return: dump
        :rtype: Dump

        """

        df_spans = self.dump_from_s_att(s_att)

        # restrict to certain values
        if len(values) > 0:
            if s_att not in df_spans.columns:
                logger.error("cannot restrict spans without annotation")
                df_spans = DataFrame(columns=['match', 'matchend']).set_index(
                    ['match', 'matchend']
                )
            else:
                values = set(values)
                logger.info("restricting spans using %d values" % len(values))
                df_spans = df_spans.loc[df_spans[s_att].isin(values)]

        # save as NQR
        if name is not None:
            # undump the dump and save to disk
            cqp = self.start_cqp()
            cqp.nqr_from_dump(df_spans, name)
            cqp.nqr_save(self.corpus_name, name)
            cqp.__kill__()

        # return proper Dump
        return Dump(self.copy(), df_spans, name_cqp=None)

    def query(self, cqp_query, context=20, context_left=None,
              context_right=None, context_break=None, corrections=dict(),
              match_strategy='standard', name=None):
        """Get query result as (context-extended) Dump (with corrected
        anchors). If a name is given, the resulting NQR (without
        context and before anchor correction) will be written to disk
        in CWB binary format.

        === (match, matchend), 0*, ..., 9*, contextid*, context*, contextend* ===

        :param str query: CQP query
        :param int context: maximum context around match..matchend (symmetric)
        :param int context_left: maximum context left to the match
        :param int context_right: maximum context right to the matchend
        :param str context_break: s-attribute to confine context to
        :param dict corrections: corrections {anchor: offset}
        :param str match_strategy: CQP matching strategy
        :param str name: name for NQR

        :return: dump
        :rtype: Dump

        """

        # preprocess input
        save = False if name is None else True  # save NQR from CQP to disk?
        name = 'Last' if name is None else name  # name in CQP
        query, s_query, anchors = preprocess_query(cqp_query)
        s_query = context_break if s_query is None else s_query
        context_left = context if context_left is None else context_left
        context_right = context if context_right is None else context_right

        # get dump from query
        df_dump = self.dump_from_query(
            query=query,
            s_query=s_query,
            anchors=anchors,
            match_strategy=match_strategy,
            name=name,
            save=save
        )

        # if dump has been retrieved from cache, NQR might not exist
        if self.show_nqr().empty or \
           name not in self.show_nqr()['subcorpus'].values:
            # undump the dump and save to disk
            cqp = self.start_cqp()
            cqp.nqr_from_dump(df_dump, name)
            cqp.nqr_save(self.corpus_name, name)
            cqp.__kill__()

        # empty return?
        if len(df_dump) == 0:
            logger.warning("found 0 matches")
            df_dump = DataFrame(columns=['match', 'matchend']).set_index(
                ['match', 'matchend']
            )
        else:
            # extend dump to context
            df_dump = self.dump2context(
                df_dump, context_left, context_right, context_break
            )
            # apply corrections to anchor points
            df_dump = correct_anchors(df_dump, corrections)

        # return proper dump
        return Dump(
            self.copy(),
            df_dump,
            name_cqp=name
        )



from ccc import Corpus
from ccc import Corpora
import pandas as pd
import pytest

from .conftest import LOCAL, DATA_PATH


def get_corpus(corpus_settings, data_path=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_path=corpus_settings['registry_path'],
        lib_path=corpus_settings.get('lib_path', None),
        data_path=data_path
    )


###############################################
# INIT ########################################
###############################################

@pytest.mark.init
def test_corpora(germaparl):
    corpora = Corpora(registry_path=germaparl['registry_path'])
    assert(type(corpora.show()) == pd.DataFrame)
    assert("GERMAPARL1386" in corpora.show().index)


@pytest.mark.init
def test_corpus(germaparl):
    corpus = get_corpus(germaparl)
    assert(corpus.corpus_size > 1000)


@pytest.mark.init
def test_get_corpus(germaparl):
    corpus = get_corpus(germaparl)
    assert(corpus.corpus_size > 1000)


@pytest.mark.init
def test_corpus_descriptor(germaparl):
    corpus = get_corpus(germaparl)
    assert(isinstance(corpus.attributes_available, pd.DataFrame))


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.init
def test_corpus_lib(brexit):
    corpus = get_corpus(brexit)
    assert(corpus.corpus_size > 1000)


#####################################################
# ATTRIBUTES ########################################
#####################################################

@pytest.mark.attributes
def test_cpos2patt(germaparl):
    corpus = get_corpus(germaparl)
    token = corpus.cpos2patts(124345)
    assert(isinstance(token, tuple))
    assert(token[0] == 'gilt')


@pytest.mark.attributes
def test_cpos2patts(germaparl):
    corpus = get_corpus(germaparl)
    token = corpus.cpos2patts(124345, ['word', 'pos'])
    assert(isinstance(token, tuple))
    assert(token == ('gilt', 'VVFIN'))


@pytest.mark.attributes
def test_marginals_word(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals(["Merkel", "Seehofer", "gehen"])
    print(df)


@pytest.mark.attributes
def test_marginals_lemma(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals(["Merkel", "Seehofer", "gehen"], p_att='lemma')
    print(df)


@pytest.mark.attributes
def test_marginals_lemma_pos(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals_complex([('Seehofer', 'NE')], ['lemma', 'pos'])
    print(df)


@pytest.mark.attributes
def test_marginals_pattern(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals(["Merkel", "Seehofer", "geh.*"], pattern=True)
    print(df)


@pytest.mark.attributes
def test_marginals_complex(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals_complex(
        [("gehen", "VVFIN"), ("Seehofer", "NE"), ("Merkel", "NE")],
        ["lemma", "pos"]
    )
    print(df)


#####################################################
# SUBCORPORA ########################################
#####################################################

@pytest.mark.subcorpus
def test_activate_subcorpus(germaparl):

    corpus = get_corpus(germaparl)

    df1 = corpus.dump_from_query(
        "[lemma='Seehofer']"
    )
    corpus.query(
        "[lemma='Horst'] expand to s",
        name='Horst'
    ).df
    corpus.activate_subcorpus('Horst')
    df2 = corpus.dump_from_query(
        "[lemma='Seehofer']"
    )
    assert(len(df1) > len(df2))


@pytest.mark.subcorpus
def test_deactivate_subcorpus(germaparl):

    corpus = get_corpus(germaparl)

    df1 = corpus.dump_from_query(
        '[lemma="die"]',
        germaparl['s_query']
    )

    # define subcorpus
    corpus.query(
        cqp_query="[lemma='sein'] expand to s",
        name='Sein'
    )

    # activate subcorpus
    corpus.activate_subcorpus('Sein')
    df2 = corpus.dump_from_query(
        '[lemma="die"]',
        germaparl['s_query']
    )

    # deactivate subcorpus
    corpus.activate_subcorpus()
    df3 = corpus.dump_from_query(
        '[lemma="die"]',
        germaparl['s_query']
    )

    assert(len(df1) == len(df3))
    assert(len(df1) > len(df2))


@pytest.mark.subcorpus
def test_create_cached_nqr(germaparl):

    # problem: if a query runs once without having been given a name,
    # the CQP dump (NQR) is not saved to disk
    # if the same query runs again *with a name*, it should be saved to disk

    corpus = get_corpus(germaparl)
    corpus.query('[lemma="jetzt"]')
    corpus.query('[lemma="jetzt"]', name='Jetzt')
    corpus.activate_subcorpus("Jetzt")
    assert("Jetzt" in corpus.show_nqr().values)


def test_nqr_from_s_att(germaparl):

    corpus = get_corpus(germaparl)
    corpus.query_s_att("text_party", values={"CDU", "CSU"}, name="Union")
    corpus.activate_subcorpus("Union")


################################################
# CREATING DUMPS ###############################
################################################


@pytest.mark.dump
def test_dump_from_s_att(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.dump_from_s_att('text_id')
    assert(df.iloc[0]['text_id'] == "i13_86_1_1")


@pytest.mark.dump
def test_dump_from_s_att_wo(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.dump_from_s_att('p')
    assert(df.shape[0] == 7332)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.dump
def test_dump_from_s_att_with(brexit):
    corpus = get_corpus(brexit)
    df = corpus.dump_from_s_att('ner_type')
    assert(len(df) == 1212944)


@pytest.mark.dump
def test_dump_from_query(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    assert(isinstance(df_dump, pd.DataFrame))
    assert(df_dump.shape[0] == 30)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.dump
def test_dump_from_query_1(brexit):
    corpus = get_corpus(brexit)
    df_dump = corpus.dump_from_query(
        query='[lemma="angela"] @1[lemma="merkel"]',
        anchors=[1],
        match_strategy='longest'
    )
    assert(isinstance(df_dump, pd.DataFrame))
    assert(df_dump.shape[0] > 99)


@pytest.mark.dump
def test_dump_from_query_anchors(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query_anchor'],
        s_query=germaparl['s_query'],
        anchors=germaparl['anchors'],
        match_strategy='standard'
    )
    assert(isinstance(df_dump, pd.DataFrame))
    assert(df_dump.shape[0] == 30)
    assert(all(elem in df_dump.columns for elem in germaparl['anchors']))


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.dump
def test_dump_from_query_lib(brexit):
    corpus = get_corpus(brexit)
    df_dump = corpus.dump_from_query(
        query=brexit['query_lib'],
        s_query=brexit['s_query'],
        match_strategy='longest'
    )
    assert(isinstance(df_dump, pd.DataFrame))
    assert(df_dump.shape[0] > 99)


#################################################
# WORKING ON DUMPS ##############################
#################################################

@pytest.mark.dumpp
def test_dump2satt(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump['test'] = None
    df = corpus.dump2satt(df_dump, germaparl['s_meta'])
    assert('test' in df.columns)
    assert(len(df) == 30)
    assert(df.iloc[0]['text_id_span'] == 10628)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.dumpp
def test_dump2satt_2(brexit):
    corpus = get_corpus(brexit)
    df_dump = corpus.dump_from_query(
        query=brexit['query'],
        s_query=brexit['s_query'],
        match_strategy='longest'
    )
    df = corpus.dump2satt(df_dump, 'vp')
    assert(df.iloc[0]['vp_cwbid'] == 112)


@pytest.mark.dumpp
def test_dump2context(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 's')
    assert(all(
        elem in df_dump.columns for elem in ['context', 'contextid', 'contextend']
    ))


@pytest.mark.dumpp
def test_dump2context2(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 'text_id')
    assert(all(elem in df_dump.columns for elem in [
        'context', 'contextid', 'contextend', 'text_id_cwbid', 'text_id'
    ]))


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.dumpp
def test_dump2context3(brexit):
    corpus = get_corpus(brexit)
    df_dump = corpus.dump_from_query(
        query=brexit['query'],
        s_query=brexit['s_query'],
        match_strategy='longest'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 'np')
    assert(all(elem in df_dump.columns for elem in [
        'context', 'contextid', 'contextend', 'np_cwbid'
    ]))


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.dumpp
def test_dump2context4(brexit):
    corpus = get_corpus(brexit)
    df_dump = corpus.dump_from_query(
        query=brexit['query'],
        s_query=brexit['s_query'],
        match_strategy='longest'
    )
    df_dump = corpus.dump2context(df_dump, None, 5, 'tweet_id')
    assert(all(elem in df_dump.columns for elem in [
        'context', 'contextid', 'contextend', 'tweet_id_cwbid', 'tweet_id'
    ]))


@pytest.mark.dumpp
def test_dump2patt(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    # df = corpus.dump2context(df_dump, 20, 20, 'text_id')
    df_dump = corpus.dump2patt(df_dump)
    assert('word' in df_dump.columns)


@pytest.mark.dumpp
def test_dump2patt2(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 'text_id')
    df_dump = corpus.dump2patt(df_dump, start='context', end='contextend')
    assert('word' in df_dump.columns)


#################################################
# QUERY ALIASES #################################
#################################################

@pytest.mark.query
def test_query_context_1(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context=None
    ).df
    assert(isinstance(df, pd.DataFrame))
    columns = germaparl['anchors'] + ['context', 'contextend']
    assert(all(elem in df.columns for elem in columns))


@pytest.mark.query
def test_query_context_2(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context_left=10,
        context=15
    ).df
    assert(isinstance(df, pd.DataFrame))
    columns = germaparl['anchors'] + ['context', 'contextend']
    assert(all(elem in df.columns for elem in columns))


@pytest.mark.query
def test_query_context_3(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context=None,
        context_break='s'
    ).df
    assert(isinstance(df, pd.DataFrame))
    columns = germaparl['anchors'] + ['contextid', 'context', 'contextend']
    assert(all(elem in df.columns for elem in columns))


@pytest.mark.query
def test_query_context_4(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context=10,
        context_break='s'
    ).df
    assert(isinstance(df, pd.DataFrame))
    columns = germaparl['anchors'] + ['contextid', 'context', 'contextend']
    assert(all(elem in df.columns for elem in columns))


@pytest.mark.query
def test_query_context_5(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query='[lemma="Horst"]',
        context=10,
        context_break='s'
    ).df
    assert(isinstance(df, pd.DataFrame))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# part of module
from ccc.utils import formulate_cqp_query, calculate_offset
from ccc.collocates import df_node_to_cooc, add_ams
from ccc.concordances import Concordance
from ccc.collocates import Collocates
# requirements
from pandas import merge, DataFrame
# logging
from .utils import time_it
import logging
logger = logging.getLogger(__name__)


class Discourseme:
    """
    realization of a discourseme with context in a corpus
    """

    def __init__(self, corpus, items, p_query, s_query, flags="%cd", escape=False):
        """
        .nodes
        .idx
        .dump
        """

        self.nodes = {
            'items': items,
            'p_query': p_query,
            's_query': s_query,
            'flags': flags,
            'escape': escape,
            'corpus': corpus.corpus_name
        }
        self.query = formulate_cqp_query(items, p_query, s_query, flags, escape)
        self.idx = corpus.cache.generate_idx(self.nodes.values(), prefix='nodes_')
        self.dump = corpus.query(self.query, context=None, context_break=s_query)

    def concordance(self, context=None, context_left=None,
                    context_right=None, context_break=None,
                    p_show=['word'], s_show=[], order='random',
                    cut_off=100, matches=None, form='dataframe'):

        self.dump.set_context(context, context_break,
                              context_left, context_right)

        return self.dump.concordance(
            matches=matches, p_show=p_show, s_show=s_show,
            slots=[], order=order,
            cut_off=cut_off, form=form
        )

    def collocates(self, window_sizes=[3, 5, 7], context_break=None,
                   order='log_likelihood', cut_off=100,
                   p_query="lemma", ams=None, min_freq=2,
                   frequencies=True, flags=None):

        # determine mode and mws
        if type(window_sizes) is int:
            single = True
            mws = window_sizes
        elif type(window_sizes) is list:
            single = False
            mws = max(window_sizes)
        else:
            raise NotImplementedError("window_sizes must be int or list")

        self.dump.set_context(context=mws, context_break=context_break)

        coll = Collocates(
            corpus=self.dump.corpus.copy(),
            df_dump=self.dump.df,
            p_query=p_query,
            mws=mws
        )

        if single:
            return coll.show(
                window=mws, order=order, cut_off=cut_off, ams=ams,
                min_freq=min_freq, frequencies=frequencies, flags=flags
            )

        else:
            collocates = dict()
            for window in window_sizes:
                collocates[window] = coll.show(
                    window=window, order=order, cut_off=cut_off, ams=ams,
                    min_freq=min_freq, frequencies=frequencies, flags=flags
                )
            return collocates


class DiscoursemeConstellation:
    """
    realization of a discourseme constellation given a topic and discoursemes

    discoursemes can be added after initialization
    """

    def __init__(self, topic, discoursemes={}, name='Last'):
        """
        .topic
        .corpus
        .nodes
        .name

        .discoursemes: dict of Discs with key == Disc.idx
        .df_nodes = dict of df_nodes with key == window
        """

        self.topic = topic

        self.corpus = topic.dump.corpus
        self.nodes = topic.nodes
        self.name = name

        self.discoursemes = dict()
        for d in discoursemes:
            self.add_disc(d)

        self.df_nodes = dict()

    def add_disc(self, disc):

        if disc.idx in self.discoursemes.keys():
            logger.warning("discourseme already included in constellation")
        else:
            self.discoursemes[disc.idx] = disc
            self.df_nodes = dict()  # reset df_nodes

    def add_items(self, items):

        disc = Discourseme(
            corpus=self.corpus.copy(),
            items=items,
            **{i: self.nodes[i] for i in self.nodes if i not in ['corpus', 'items']}
        )
        self.add_disc(disc)
        return disc.idx

    @time_it
    def slice_discs(self, window=5, context_break=None):
        """ intersects all discoursemes
        :return: df_nodes indexed by topic-matches:

        == (m, m-end) c-id, c, c-end, m_d1, m-e_d1, ... ==

        with duplicates for each match, matchend where necessary
        """

        topic_dump = self.topic.dump
        topic_dump.set_context(context=window, context_break=context_break)
        df_nodes = topic_dump.df.reset_index()

        for disc in self.discoursemes.values():

            df = disc.dump.df.reset_index()
            df = df.drop(['context', 'contextend'], axis=1)
            # merge nodes; NB: this adds duplicates where necessary
            # s_break = self.topic.nodes['s_query'] + '_cwbid'
            s_break = 'contextid'
            df_nodes = merge(df_nodes, df, on=s_break)
            # remove lines where items are too far away
            df_nodes['offset'] = df_nodes.apply(calculate_offset, axis=1)
            df_nodes = df_nodes[
                abs(df_nodes['offset']) <= window
            ]
            df_nodes = df_nodes.rename(columns={
                'match_x': 'match',
                'matchend_x': 'matchend',
                'match_y': 'match_' + disc.idx,
                'matchend_y': 'matchend_' + disc.idx,
                'offset': 'offset_' + disc.idx
            })

        self.df_nodes[window] = df_nodes

        return df_nodes

    @time_it
    def concordance(self, window=5, matches=None,
                    p_show=['word'], s_show=[], order='random',
                    cut_off=100, form='dataframe'):

        """ self.df_nodes has duplicate entries
        (1) convert to (match matchend) disc_1_set disc_2_set ...
        (2) convert each line to dataframe
        """

        # make sure we're having the right context
        if window not in self.df_nodes.keys():
            df_nodes = self.slice_discs(window).copy()
        else:
            df_nodes = self.df_nodes[window]

        # get ids of all discoursemes
        disc_ids = set(self.discoursemes.keys())

        logger.info("converting discourse nodes to regular dump")
        # TODO speed up
        all_matches = set(df_nodes['match'])
        rows = list()
        for match in all_matches:
            row = dict()
            df_loc = df_nodes.loc[df_nodes.match == match]
            row['match'] = match
            row['matchend'] = df_loc.iloc[0]['matchend']
            row['context id'] = df_loc.iloc[0]['contextid']
            row['context'] = df_loc.iloc[0]['context']
            row['contextend'] = df_loc.iloc[0]['contextend']
            for idx in disc_ids:
                disc_f1 = set()
                for a, b in zip(df_loc['match_' + idx], df_loc['matchend_' + idx]):
                    disc_f1.update(range(a, b + 1))
                row[idx] = disc_f1
            rows.append(row)
        df = DataFrame(rows)
        df = df.set_index(["match", "matchend"])

        logger.info("converting each line to dataframe")
        conc = Concordance(self.corpus.copy(), df)
        lines = conc.lines(
            matches=matches, p_show=p_show, s_show=s_show,
            slots=[], order=order,
            cut_off=cut_off, form=form
        )

        logger.info("inserting discourseme and window/context info")
        # TODO mark out of context
        dfs = list()
        for line in lines.iterrows():
            df = line[1]['dataframe']
            # indicate topic matches
            match, matchend = line[0]
            df[self.topic.idx] = df.index.isin(set(range(match, matchend + 1)))
            # indicate discourseme matches
            for idx in disc_ids:
                df[idx] = df.index.isin(line[1][idx])
            # df = df.drop(['match', 'matchend', 'context', 'contextend'], axis=1)
            dfs.append(df)
        lines['dataframe'] = dfs

        return lines

    @time_it
    def collocates(self, window=5, order='log_likelihood', cut_off=100,
                   p_query="lemma", ams=None, min_freq=2,
                   frequencies=True, flags=None):

        # make sure we're having the right context
        if window not in self.df_nodes.keys():
            df_nodes = self.slice_discs(window)
        else:
            df_nodes = self.df_nodes[window]

        # get and correct df_cooc, f1_set
        df_cooc, f1_set = df_node_to_cooc(df_nodes)

        if len(f1_set) == 0:
            logger.warning("no matches")
            return DataFrame()

        # check for presence of topic in topic context
        # NB: this is necessary since we might have excluded
        # some occurrences of the topic from df_nodes
        # when selecting relevant instances in the constellation
        logger.info('searching for topic discourseme in topic context')
        f1_set = self.topic.dump.matches()

        # check for presence of discoursemes in topic context
        for idx in self.discoursemes:
            logger.info('searching for discourseme "%s" in topic context' % idx)
            matches = self.discoursemes[idx].dump.matches()
            f1_set.update(matches)

        logger.info("excluding all discourseme matches from context")
        df_cooc = df_cooc.loc[~df_cooc['cpos'].isin(f1_set)]

        # moving to requested window
        relevant = df_cooc.loc[abs(df_cooc['offset']) <= window]

        # number of possible occurrence positions within window
        f1_inflated = len(relevant)

        # get frequency counts
        f = self.corpus.counts.cpos(
            relevant['cpos'], [p_query]
        )
        f.columns = ['f']
        f.index = f.index.get_level_values(p_query)

        # get marginals
        f2 = self.corpus.marginals(
            f.index, p_query
        )
        f2.columns = ['marginal']

        # deduct node frequencies from marginals
        logger.info("deducting node frequencies")
        node_freq = self.corpus.counts.cpos(f1_set, [p_query])
        node_freq.index = node_freq.index.get_level_values(p_query)
        node_freq.columns = ['in_nodes']
        f2 = f2.join(node_freq)
        f2 = f2.fillna(0)
        f2['in_nodes'] = f2['in_nodes'].astype(int)
        f2['f2'] = f2['marginal'] - f2['in_nodes']

        # get sub-corpus size
        f1 = f1_inflated

        # get corpus size
        N = self.corpus.corpus_size - len(f1_set)

        collocates = add_ams(
            f, f1, f2, N,
            min_freq, order, cut_off, flags, ams, frequencies
        )

        return collocates

    # def collocates_range(self, windows=[3, 5, 7, 10], order='f', cut_off=100,
    #                      p_query="lemma", ams=None, min_freq=2,
    #                      frequencies=True, flags=None):

    #     # sort windows and start with largest
    #     windows = sorted(windows, reverse=True)
    #     max_window = windows[0]

    #     # make sure we're having the right context
    #     if max_window not in self.df_nodes.keys():
    #         df_nodes = self.slice_discs(max_window)
    #     else:
    #         df_nodes = self.df_nodes[max_window]

    #     # get and correct df_cooc, f1_set
    #     df_cooc, f1_set = df_node_to_cooc(df_nodes)

    #     if len(f1_set) == 0:
    #         logger.warning("no matches")
    #         return DataFrame()

    #     # check for presence of topic in topic context
    #     # NB: this is necessary since we might have excluded
    #     # some occurrences of the topic from df_nodes
    #     # when selecting relevant instances in the constellation
    #     logger.info('searching for topic discourseme in topic context')
    #     f1_set = self.topic.dump.matches()

    #     # check for discoursemes in topic context
    #     for idx in self.discoursemes:
    #         logger.info('searching for discourseme "%s" in topic context' % idx)
    #         matches = self.discoursemes[idx].dump.matches()
    #         f1_set.update(matches)
    #         df_cooc = df_cooc.loc[~df_cooc['cpos'].isin(f1_set)]

    #     logger.info("excluding all discourseme matches from context")
    #     df_cooc = df_cooc.loc[~df_cooc['cpos'].isin(f1_set)]

    #     # do it once for max_window
    #     relevant = df_cooc.loc[abs(df_cooc['offset']) <= max_window]

    #     # number of possible occurrence positions within window
    #     f1_inflated = len(relevant)

    #     # get frequency counts
    #     f = self.corpus.counts.cpos(
    #         relevant['cpos'], [p_query]
    #     )
    #     f.columns = ['f']
    #     f.index = f.index.get_level_values(p_query)

    #     # get marginals
    #     f2 = self.corpus.marginals(
    #         f.index, p_query
    #     )
    #     f2.columns = ['marginal']

    #     # deduct node frequencies from marginals
    #     logger.info("deducting node frequencies")
    #     node_freq = self.corpus.counts.cpos(f1_set, [p_query])
    #     node_freq.index = node_freq.index.get_level_values(p_query)
    #     node_freq.columns = ['in_nodes']
    #     f2 = f2.join(node_freq)
    #     f2 = f2.fillna(0)
    #     f2['in_nodes'] = f2['in_nodes'].astype(int)
    #     f2['f2'] = f2['marginal'] - f2['in_nodes']

    #     # get sub-corpus size
    #     f1 = f1_inflated

    #     # get corpus size
    #     N = self.corpus.corpus_size - len(f1_set)

    #     collocates = add_ams(
    #         f, f1, f2, N,
    #         min_freq, order, cut_off, flags, ams, frequencies
    #     )

    #     # repeat for all other window sizes
    #     for w in windows[1:]:
    #         print(w)

    #     return collocates


# class DiscoursemeConstellation:

#     def __init__(self, topic, discoursemes={}, name='Last'):
#         """
#         :param DataFrame topic.dump.df: (match, matchend) context contextend contextid
#         :param Corpus topic.corpus: corpus with subcorpus = topic
#         :
#         .discoursemes: dict of Discs with key == Disc.idx
#         .df_nodes = dict of df_nodes with key == window
#         """

#         self.topic = topic

#         self.corpus = topic.dump.corpus
#         self.nodes = topic.nodes
#         self.name = name

#         self.discoursemes = dict()
#         for d in discoursemes:
#             self.add_disc(d)

#         self.df_nodes = dict()



from ccc.cwb import Corpus
from ccc.counts import cwb_scan_corpus
from ccc.utils import formulate_cqp_query
import pandas as pd
import pytest
from tempfile import NamedTemporaryFile

from .conftest import LOCAL, DATA_PATH


def get_corpus(corpus_settings, data_path=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_path=corpus_settings['registry_path'],
        lib_path=corpus_settings.get('lib_path', None),
        data_path=data_path
    )


@pytest.mark.cwb_counts
def test_cwb_scan_corpus_subcorpora(germaparl):

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()

    # get a DumpFrame
    cqp.nqr_from_query('[lemma="Helmut"]', name='Tmp')

    # run cwb-scan-corpus on dump
    with NamedTemporaryFile(mode="wt") as f:
        cqp.Exec('dump Tmp > "%s"' % f.name)
        df1 = cwb_scan_corpus(f.name, germaparl['corpus_name'],
                              germaparl['registry_path'])

    # activate a sub-corpus
    cqp.nqr_from_query(query='[lemma="und"] expand to s', name='und')
    cqp.nqr_activate(corpus.corpus_name, 'und')

    # run same query on sub-corpus
    cqp.nqr_from_query('[lemma="Helmut"]', name='Tmp')

    # run cwb-scan-corpus on dump
    with NamedTemporaryFile(mode="wt") as f:
        cqp.Exec('dump Tmp > "%s"' % f.name)
        df2 = cwb_scan_corpus(f.name, germaparl['corpus_name'],
                              germaparl['registry_path'])

    # check that results are different
    assert(sum(df2['freq']) != sum(df1['freq']))

    cqp.__kill__()


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_cwb_scan_corpus_marginal(germaparl):

    df1 = cwb_scan_corpus(
        None,
        germaparl['corpus_name'],
        germaparl['registry_path']
    )
    df2 = cwb_scan_corpus(
        None,
        germaparl['corpus_name'],
        germaparl['registry_path'],
        p_atts=['lemma', 'pos']
    )
    assert(isinstance(df1.index, pd.MultiIndex))
    assert(isinstance(df2.index, pd.MultiIndex))
    assert(df1.index.names == ['word'])
    assert(df2.index.names == ['lemma', 'pos'])
    assert(df1.columns == df2.columns)


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_marginals(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.marginals(["Helmut", "Kohl", "CDU"])
    assert(len(freqframe) == 3)
    assert(isinstance(freqframe, pd.DataFrame))
    print(freqframe.index)


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_marginals_patterns(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.marginals(["H.*", "Kohl", "CDU"])
    assert(len(freqframe) == 3)
    assert(isinstance(freqframe, pd.DataFrame))
    assert(freqframe.loc['H.*']['freq'] == 0)
    freqframe = corpus.marginals(["H.*", "Kohl", "CDU"], pattern=True)
    assert(len(freqframe) == 3)
    assert(isinstance(freqframe, pd.DataFrame))
    assert(freqframe.loc['H.*']['freq'] == 2466)


@pytest.mark.cwb_counts
def test_count_cpos(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.counts.cpos(list(range(1, 1000)), p_atts=['word'])
    assert(isinstance(freqframe, pd.DataFrame))
    assert(isinstance(freqframe.index, pd.MultiIndex))


@pytest.mark.cwb_counts
def test_count_cpos_combo(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.counts.cpos(list(range(1, 1000)), p_atts=['lemma', 'pos'])
    assert(isinstance(freqframe, pd.DataFrame))
    assert(isinstance(freqframe.index, pd.MultiIndex))


@pytest.mark.cwb_counts
def test_count_items(germaparl):

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()

    items = ["Helmut", "Kohl", "CDU"]
    queries = [
        formulate_cqp_query([item]) for item in items
    ]

    # whole corpus
    counts1 = corpus.marginals(items)
    counts2 = corpus.counts.mwus(cqp, queries)
    print(counts1, counts2)
    assert(list(counts1["freq"]) == list(counts2["freq"]))

    # subcorpus
    cqp.nqr_from_query(query='[lemma="und"] expand to s', name='und')
    cqp.nqr_activate(corpus.corpus_name, 'und')
    counts1 = corpus.marginals(items)
    counts2 = corpus.counts.mwus(cqp, queries)
    assert(counts1.loc[items[0], 'freq'] > counts2.loc[queries[0], 'freq'])

    # whole corpus
    cqp.nqr_activate(corpus.corpus_name)
    counts1 = corpus.marginals(items)
    counts2 = corpus.counts.mwus(cqp, queries)
    assert(list(counts1["freq"]) == list(counts2["freq"]))

    cqp.__kill__()


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
@pytest.mark.cwb_counts
def test_count_matches(brexit):
    corpus = get_corpus(brexit)
    corpus.query(
        cqp_query='[lemma="nigel"]',
        context=10,
        context_break='tweet',
        name='Nigel'
    )
    cqp = corpus.start_cqp()
    counts = corpus.counts.matches(cqp, 'Nigel')
    cqp.__kill__()
    assert("Nigel" in counts.index)


@pytest.mark.mwus
def test_count_mwus_3(germaparl):

    # whole corpus
    corpus = get_corpus(germaparl)
    items = ["Horst Seehofer", r"( CSU )", "CSU", "WES324", "CSU"]
    queries = [
        formulate_cqp_query([item]) for item in items
    ]

    cqp = corpus.start_cqp()
    counts3 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=3,
        fill_missing=False
    )
    cqp.__kill__()

    assert(counts3['freq']['CSU'] == 635)


@pytest.mark.mwus
@pytest.mark.cwb_counts
def test_count_mwus_strategies(germaparl):

    # whole corpus
    corpus = get_corpus(germaparl)
    items = ["Horst Seehofer", r"( CSU )", "CSU", "WES324"]
    queries = [
        formulate_cqp_query([item]) for item in items
    ]

    cqp = corpus.start_cqp()
    counts1 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=1,
        fill_missing=False
    )
    assert('([word="CSU"])' in counts1.index)

    counts2 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=2,
        fill_missing=False
    )

    counts3 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=3,
        fill_missing=False
    )

    cqp.__kill__()
    assert(counts2.equals(counts3))
    assert(sum(counts1['freq']) == sum(counts2['freq']))


@pytest.mark.cwb_counts
def test_count_items_subcorpora(germaparl):

    # subcorpus
    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    dump = corpus.dump_from_s_att("text_role", ["presidency"])
    cqp.nqr_from_dump(dump, 'Presidency')
    cqp.nqr_activate(corpus.corpus_name, 'Presidency')
    items = ["Horst Seehofer", r"( CSU )", "CSU", "WES324", "CSU"]
    queries = [formulate_cqp_query([item]) for item in items]

    counts1 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=1,
        fill_missing=False
    )
    assert(sum(counts1['freq']) > 0)

    counts2 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=2,
        fill_missing=False
    )

    counts3 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=3,
        fill_missing=False
    )
    assert(counts2.equals(counts3))
    cqp.__kill__()


# def test_count_dump_1(germaparl):

#     strategy = 1
#     corpus = get_corpus(germaparl)
#     df_dump = corpus.dump_from_query(
#         query=germaparl['query'],
#         s_query=germaparl['s_query'],
#         match_strategy='standard'
#     )

#     # no split
#     # - easy
#     counts_ns_e = corpus.count_dump(df_dump,
#                                     strategy=strategy,
#                                     split=False,
#                                     p_atts=['word'])
#     # - combo
#     counts_ns_c = corpus.count_dump(df_dump,
#                                     strategy=strategy,
#                                     split=False,
#                                     p_atts=['word', 'pos'])

#     # split
#     # - easy
#     counts_s_e = corpus.count_dump(df_dump,
#                                    strategy=strategy,
#                                    split=True,
#                                    p_atts=['word'])
#     # - combo
#     counts_s_c = corpus.count_dump(df_dump,
#                                    strategy=strategy,
#                                    split=True,
#                                    p_atts=['word', 'pos'])

#     print(counts_ns_e)
#     print(counts_ns_c)
#     print(counts_s_e)
#     print(counts_s_c)


# def test_count_dump_2(germaparl):

#     strategy = 2
#     corpus = get_corpus(germaparl)
#     df_dump = corpus.dump_from_query(
#         query=germaparl['query'],
#         s_query=germaparl['s_query'],
#         match_strategy='standard'
#     )

#     # no split
#     # - easy
#     counts_ns_e = corpus.count_dump(df_dump,
#                                     strategy=strategy,
#                                     split=False,
#                                     p_atts=['word'])
#     # - combo
#     counts_ns_c = corpus.count_dump(df_dump,
#                                     strategy=strategy,
#                                     split=False,
#                                     p_atts=['word', 'pos'])

#     # split
#     # - easy
#     counts_s_e = corpus.count_dump(df_dump,
#                                    strategy=strategy,
#                                    split=True,
#                                    p_atts=['word'])
#     # - combo
#     counts_s_c = corpus.count_dump(df_dump,
#                                    strategy=strategy,
#                                    split=True,
#                                    p_atts=['word', 'pos'])

#     print(counts_ns_e)
#     print(counts_ns_c)
#     print(counts_s_e)
#     print(counts_s_c)


@pytest.mark.cwb_counts
def test_counts_dump_1_split(germaparl):
    strategy = 1

    corpus = get_corpus(germaparl)
    dump = corpus.dump_from_query('[lemma="die" %cd] [pos="N.*"]')

    df = corpus.counts.dump(dump, p_atts=['word'], split=True, strategy=strategy)
    assert(int(df["freq"]["der"]) == 3775)

    df = corpus.counts.dump(dump, p_atts=['word', 'lemma'], split=True, strategy=strategy)
    assert(int(df["freq"][("der", "die")]) == 3775)


@pytest.mark.cwb_counts
def test_counts_dump_1_no_split(germaparl):
    strategy = 1

    corpus = get_corpus(germaparl)
    dump = corpus.dump_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]')

    # no split
    df = corpus.counts.dump(dump, p_atts=['word'], split=False, strategy=strategy)
    assert("Helmut Kohl" in df.index)

    df = corpus.counts.dump(dump, p_atts=['word', 'pos'], split=False, strategy=strategy)
    assert(("Helmut Kohl", "NE NE") in df.index)


@pytest.mark.cwb_counts
def test_counts_dump_2(germaparl):
    strategy = 2

    corpus = get_corpus(germaparl)
    dump = corpus.dump_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]')

    df = corpus.counts.dump(dump, p_atts=['word'], split=True, strategy=strategy)
    assert(df["freq"]["Helmut"] == 6)

    df = corpus.counts.dump(dump, p_atts=['word', 'pos'], split=True, strategy=strategy)
    assert(df["freq"][("Helmut", "NE")] == 6)

    df = corpus.counts.dump(dump, p_atts=['word'], split=False, strategy=strategy)
    assert("Helmut Kohl" in df.index)
    assert(df["freq"].iloc[0] == 6)

    df = corpus.counts.dump(dump, p_atts=['word', 'pos'], split=False, strategy=strategy)
    assert(("Helmut Kohl", "NE NE") in df.index)
    assert(df["freq"].iloc[0] == 6)


@pytest.mark.cwb_counts
def test_counts_matches_1(germaparl):
    strategy = 1

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    cqp.nqr_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]', name='Last')
    cqp.nqr_activate(corpus.corpus_name, 'Last')

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               flags="%cd", strategy=strategy)
    assert("helmut kohl" in df.index)
    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               strategy=strategy)
    assert("Helmut Kohl" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               flags="%cd", strategy=strategy)
    assert("helmut" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               strategy=strategy)
    assert("Helmut" in df.index)

    cqp.__kill__()


@pytest.mark.cwb_counts
def test_counts_matches_2(germaparl):
    strategy = 2

    corpus = get_corpus(germaparl)

    cqp = corpus.start_cqp()
    cqp.nqr_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]', name='Last')
    cqp.nqr_activate(corpus.corpus_name, 'Last')
    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               flags="%cd", strategy=strategy)
    assert("helmut kohl" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               strategy=strategy)
    assert("Helmut Kohl" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               flags="%cd", strategy=strategy)
    assert("helmut" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               strategy=strategy)
    assert("Helmut" in df.index)
    cqp.__kill__()


@pytest.mark.cwb_counts
def test_counts_matches_3(germaparl):
    strategy = 3

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    cqp.nqr_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]', name='Last')
    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               flags="%cd", strategy=strategy)
    assert("helmut" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               strategy=strategy)
    assert("Helmut" in df.index)

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word', 'pos'], split=True,
                               strategy=strategy)
    assert(("Helmut", "NE") in df.index)
    cqp.__kill__()


@pytest.mark.cwb_counts
def test_counts_mwus(germaparl):
    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    queries = ['[lemma="Helmut"%cd & pos="NE"] [lemma="Kohl"]', '[lemma="Horst"]']
    df = corpus.counts.mwus(cqp,
                            queries,
                            strategy=1)
    assert(df['freq'][queries[0]] == 6)

    df = corpus.counts.mwus(cqp,
                            queries,
                            strategy=3,
                            p_atts=['lemma', 'pos'])
    assert(df['freq'][('Horst', 'NE')] == 55)

    df = corpus.counts.mwus(cqp,
                            queries,
                            strategy=3,
                            p_atts=['lemma'])

    assert(df['freq']['Horst'] == 55)
    cqp.__kill__()


@pytest.mark.cwb_counts
def test_cwb_counts(germaparl):
    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    queries = ['[lemma="Helmut"%cd & pos="NE"] [lemma="Kohl"]', '[lemma="Horst"]']
    df = corpus.counts.mwus(cqp,
                            queries)
    assert(df['freq'][queries[1]] == 55)
    cqp.__kill__()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# part of module
from .concordances import Concordance
from .collocates import Collocates
from .keywords import Keywords

from .utils import merge_intervals, correct_anchors
# requirements
from pandas import DataFrame
# logging
import logging
logger = logging.getLogger(__name__)


class Dump:
    """ result of a query """

    def __init__(self, corpus, df_dump, name_cqp):

        self.df = df_dump
        self.corpus = corpus
        self.name_cqp = name_cqp
        self.size = len(self.df)

        self._breakdown = None
        self._matches = None
        self._context = None

    def __str__(self):
        desc = ['a ccc.Dump with %d matches' % self.size]
        if self.corpus.subcorpus:
            crpssbcrps = self.corpus.corpus_name + ":" + self.corpus.subcorpus
        else:
            crpssbcrps = self.corpus.corpus_name
        desc.append('- corpus "%s" (%d tokens in whole corpus)' % (
            crpssbcrps, self.corpus.corpus_size
        ))
        if self.name_cqp is not None:
            desc.append('- name in cqp: "%s"' % self.name_cqp)
        desc.append('- columns: %s' % str(list(self.df.columns)))
        return "\n".join(desc)

    def set_context(self, context=None, context_break=None,
                    context_left=None, context_right=None):
        """Set context in the dump.

        """
        # pre-process context
        if context_left is None:
            context_left = context
        if context_right is None:
            context_right = context

        # set context
        self.df = self.corpus.dump2context(
            self.df, context_left, context_right, context_break
        )

    def correct_anchors(self, corrections):
        """Correct anchors by integer offsets.

        """
        self.df = correct_anchors(self.df, corrections)

    def breakdown(self, max_matches=None):

        if self._breakdown is None:
            if max_matches is not None and self.size > max_matches:
                logger.warning(
                    'no frequency breakdown (%d matches)' % self.size
                )
                breakdown = DataFrame(
                    index=['NODE'],
                    data=[self.size],
                    columns=['freq']
                )
                breakdown.index.name = 'word'
                return breakdown
            else:
                logger.info('creating frequency breakdown')
                breakdown = self.corpus.counts.dump(
                    df_dump=self.df,
                    start='match', end='matchend',
                    p_atts=['word'], strategy=1
                )
                self._breakdown = breakdown

        return self._breakdown

    def matches(self):
        """
        :return: cpos of (match .. matchend) regions
        :rtype: set
        """
        if self._matches is None:
            f1 = set()
            for match, matchend in self.df.index:
                f1.update(range(match, matchend + 1))
            self._matches = f1

        return self._matches

    def context(self):
        """
        :return: cpos of (context .. contextend) regions including matches
        :rtype: DataFrame
        """
        if self._context is None:
            self._context = DataFrame.from_records(merge_intervals(
                self.df[['context', 'contextend']].values.tolist()
            ), columns=['context', 'contextend'])

        return self._context

    def concordance(self, form='simple', p_show=['word'], s_show=[],
                    order='first', cut_off=100, matches=None, slots=None):

        conc = Concordance(
            self.corpus.copy(),
            df_dump=self.df
        )

        return conc.lines(
            form=form,
            p_show=p_show,
            s_show=s_show,
            order=order,
            cut_off=cut_off,
            matches=matches,
            slots=slots
        )

    def collocates(self, p_query='lemma', mws=20, window=5, order='f',
                   cut_off=100, ams=None, min_freq=2,
                   frequencies=True, flags=None, marginals='corpus'):

        mws = max(mws, window)

        coll = Collocates(
            self.corpus.copy(),
            df_dump=self.df,
            p_query=p_query,
            mws=mws
        )

        return coll.show(
            window=window,
            order=order,
            cut_off=cut_off,
            ams=ams,
            min_freq=min_freq,
            frequencies=frequencies,
            flags=flags,
            marginals=marginals
        )

    def keywords(self, p_query='lemma', order='f', cut_off=100,
                 ams=None, min_freq=2, frequencies=True, flags=None):

        kw = Keywords(
            self.corpus.copy(),
            self.df,
            p_query
        )

        return kw.show(
            order=order,
            cut_off=cut_off,
            ams=ams,
            min_freq=min_freq,
            frequencies=frequencies,
            flags=flags
        )



from pandas import read_csv
from ccc import Corpus
from ccc.ufa import UFA
import pytest

from .conftest import LOCAL, DATA_PATH


def get_corpus(corpus_settings, data_path=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_path=corpus_settings['registry_path'],
        lib_path=corpus_settings.get('lib_path', None),
        data_path=data_path
    )


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
def test_keywords(brexit):

    # get s_att-split
    meta = read_csv(brexit['meta_path'], dtype=str, sep="\t", nrows=100000)
    s_values = list(meta['ymd'].value_counts().head(5).index)
    ids = dict()
    for s in s_values:
        ids[s] = set(meta.loc[meta['ymd'] == s]['id'])

    # keywords
    corpus = get_corpus(brexit)
    ufa = UFA(corpus, ids, s_att='tweet_id')
    tables = ufa.keywords()
    print(tables)


@pytest.mark.skipif(not LOCAL, reason='works on my machine')
@pytest.mark.brexit
def test_collocates(brexit):

    # get s_att-split
    meta = read_csv(brexit['meta_path'], dtype=str, sep="\t", nrows=100000)
    s_values = list(meta['ymd'].value_counts().head(5).index)
    ids = dict()
    for s in s_values:
        ids[s] = set(meta.loc[meta['ymd'] == s]['id'])

    # keywords
    corpus = get_corpus(brexit)
    ufa = UFA(corpus, ids, s_att='tweet_id')
    tables = ufa.collocates(cqp_query='[lemma="johnson"]', order='log_likelihood')
    print(tables)
